"""Multitenancy route manager."""

import logging
from typing import List, Optional

from ..connections.models.conn_record import ConnRecord
from ..core.profile import Profile
from ..messaging.responder import BaseResponder
from ..protocols.coordinate_mediation.v1_0.manager import MediationManager
from ..protocols.coordinate_mediation.v1_0.messages.keylist_update import KeylistUpdate
from ..protocols.coordinate_mediation.v1_0.models.mediation_record import MediationRecord
from ..protocols.coordinate_mediation.v1_0.normalization import (
    normalize_from_did_key,
    normalize_to_did_key,
)
from ..protocols.coordinate_mediation.v1_0.route_manager import (
    CoordinateMediationV1RouteManager,
    RouteManager,
    RoutingInfo,
)
from ..protocols.routing.v1_0.manager import RoutingManager
from ..protocols.routing.v1_0.models.route_record import RouteRecord
from ..storage.error import StorageNotFoundError
from .base import BaseMultitenantManager

LOGGER = logging.getLogger(__name__)


class MultitenantRouteManager(RouteManager):
    """Multitenancy route manager."""

    def __init__(
        self,
        root_profile: Profile,
    ):
        """Initialize multitenant route manager."""
        self.root_profile = root_profile

    async def get_base_wallet_mediator(self) -> Optional[MediationRecord]:
        """Get base wallet's default mediator."""
        return await MediationManager(self.root_profile).get_default_mediator()

    async def _route_for_key(
        self,
        profile: Profile,
        recipient_key: str,
        mediation_record: Optional[MediationRecord] = None,
        *,
        skip_if_exists: bool = False,
        replace_key: Optional[str] = None,
    ) -> Optional[KeylistUpdate]:
        wallet_id = profile.settings["wallet.id"]
        LOGGER.debug(
            "Add route record for recipient %s to wallet %s", recipient_key, wallet_id
        )
        routing_mgr = RoutingManager(self.root_profile)
        mediation_mgr = MediationManager(self.root_profile)
        # If base wallet had mediator, only notify that mediator.
        # Else, if subwallet has mediator, notify that mediator.
        base_mediation_record = await self.get_base_wallet_mediator()
        mediation_record = base_mediation_record or mediation_record

        if skip_if_exists:
            try:
                async with self.root_profile.session() as session:
                    await RouteRecord.retrieve_by_recipient_key(session, recipient_key)

                # If no error is thrown, it means there is already a record
                LOGGER.debug(
                    "Route record already exists for recipient %s to wallet %s. Skipping",
                    recipient_key,
                    wallet_id,
                )
                return None
            except StorageNotFoundError:
                LOGGER.debug(
                    "Route record does not exist for recipient %s to wallet %s. Creating",
                    recipient_key,
                    wallet_id,
                )

        await routing_mgr.create_route_record(
            recipient_key=recipient_key, internal_wallet_id=wallet_id
        )

        # External mediation
        keylist_updates = None
        if mediation_record:
            LOGGER.debug("Adding key to mediation record for recipient %s", recipient_key)
            keylist_updates = await mediation_mgr.add_key(recipient_key)
            if replace_key:
                LOGGER.debug(
                    "Replacing key %s in mediation record %s", replace_key, recipient_key
                )
                keylist_updates = await mediation_mgr.remove_key(
                    replace_key, keylist_updates
                )
            # in order to locate the correct verkey for message packing we need
            # to use the correct profile.
            # if we are using default/base mediation then we need
            # the root_profile to create the responder.
            # if sub-wallets are configuring their own mediation, then
            # we need the sub-wallet (profile) to create the responder.
            profile = self.root_profile if base_mediation_record else profile
            responder = profile.inject(BaseResponder)
            LOGGER.debug(
                "Sending keylist updates to mediator %s", mediation_record.connection_id
            )
            await responder.send(
                keylist_updates, connection_id=mediation_record.connection_id
            )

        return keylist_updates

    async def mediation_records_for_connection(
        self,
        profile: Profile,
        conn_record: ConnRecord,
        mediation_id: Optional[str] = None,
        or_default: bool = False,
    ) -> List[MediationRecord]:
        """Determine mediation records for a connection."""
        conn_specific = await super().mediation_records_for_connection(
            profile, conn_record, mediation_id, or_default
        )
        base_mediation_record = await self.get_base_wallet_mediator()
        return [
            record
            for record in (base_mediation_record, *conn_specific)
            if record is not None
        ]

    async def routing_info(
        self,
        profile: Profile,
        mediation_record: Optional[MediationRecord] = None,
    ) -> RoutingInfo:
        """Return routing info."""
        LOGGER.debug(
            "Getting routing info for profile %s", profile.settings.get("wallet.id", "")
        )
        routing_keys = []

        base_mediation_record = await self.get_base_wallet_mediator()

        my_endpoint = None
        if base_mediation_record:
            routing_keys = base_mediation_record.routing_keys
            my_endpoint = base_mediation_record.endpoint

        if mediation_record:
            routing_keys = [*routing_keys, *mediation_record.routing_keys]
            my_endpoint = mediation_record.endpoint

        routing_keys = [normalize_to_did_key(key).key_id for key in routing_keys] or None

        return RoutingInfo(routing_keys, my_endpoint)


class BaseWalletRouteManager(CoordinateMediationV1RouteManager):
    """Route manager for operations specific to the base wallet."""

    async def connection_from_recipient_key(
        self, profile: Profile, recipient_key: str
    ) -> ConnRecord:
        """Retrieve a connection by recipient key.

        The recipient key is expected to be a local key owned by this agent.

        Since the multi-tenant base wallet can receive and send keylist updates
        for sub wallets, we check the sub wallet's connections before the base
        wallet.
        """
        LOGGER.debug("Retrieving connection for recipient key for multitenant wallet")
        manager = profile.inject(BaseMultitenantManager)
        profile_to_search = (
            await manager.get_profile_for_key(
                profile.context, normalize_from_did_key(recipient_key)
            )
            or profile
        )

        return await super().connection_from_recipient_key(
            profile_to_search, recipient_key
        )



"""Represents a feature discovery disclosure message."""

from typing import Mapping, Sequence

from marshmallow import EXCLUDE, Schema, ValidationError, fields

from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from .....messaging.models.base import BaseModelError
from ..message_types import DISCLOSURES, PROTOCOL_PACKAGE

HANDLER_CLASS = f"{PROTOCOL_PACKAGE}.handlers.disclosures_handler.DisclosuresHandler"


class ProtocolOrGoalCodeDescriptorField(fields.Field):
    """ProtocolDescriptor or GoalCodeDescriptor for Marshmallow."""

    def _serialize(self, value, attr, obj, **kwargs):
        return value

    def _deserialize(self, value, attr, data, **kwargs):
        try:
            GoalCodeDescriptorSchema().load(value)
            return value
        except ValidationError:
            try:
                ProtocolDescriptorSchema().load(value)
                return value
            except ValidationError:
                raise BaseModelError(
                    "Field should be ProtocolDescriptor or GoalCodeDescriptor"
                )


class ProtocolDescriptorSchema(Schema):
    """Schema for an entry in the protocols list."""

    id = fields.Str(required=True)
    feature_type = fields.Str(
        required=True, data_key="feature-type", metadata={"description": "feature-type"}
    )
    roles = fields.List(
        fields.Str(
            metadata={
                "description": "Role: requester or responder",
                "example": "requester",
            }
        ),
        required=False,
        allow_none=True,
        metadata={"description": "List of roles"},
    )


class GoalCodeDescriptorSchema(Schema):
    """Schema for an entry in the goal_code list."""

    id = fields.Str(required=True)
    feature_type = fields.Str(
        required=True, data_key="feature-type", metadata={"description": "feature-type"}
    )


class Disclosures(AgentMessage):
    """Represents a feature discovery disclosure, the response to a query message."""

    class Meta:
        """Disclose metadata."""

        handler_class = HANDLER_CLASS
        message_type = DISCLOSURES
        schema_class = "DisclosuresSchema"

    def __init__(self, *, disclosures: Sequence[Mapping] = None, **kwargs):
        """Initialize disclose message object.

        Args:
            disclosures: A mapping of protocol names to a dictionary of properties
            kwargs: Additional keyword arguments for the message
        """
        super().__init__(**kwargs)
        self.disclosures = list(disclosures) if disclosures else []


class DisclosuresSchema(AgentMessageSchema):
    """Disclose message schema used in serialization/deserialization."""

    class Meta:
        """DiscloseSchema metadata."""

        model_class = Disclosures
        unknown = EXCLUDE

    disclosures = fields.List(
        ProtocolOrGoalCodeDescriptorField(),
        required=True,
        metadata={"description": "List of protocol or goal_code descriptors"},
    )



"""AnonCreds specific credential exchange information with non-secrets storage."""

from typing import Any, Mapping, Optional, Sequence

from marshmallow import EXCLUDE, fields

from ......core.profile import ProfileSession
from ......messaging.models.base_record import BaseRecord, BaseRecordSchema
from ......messaging.valid import (
    ANONCREDS_CRED_DEF_ID_EXAMPLE,
    ANONCREDS_REV_REG_ID_EXAMPLE,
    UUID4_EXAMPLE,
)
from .. import UNENCRYPTED_TAGS


class V20CredExRecordAnonCreds(BaseRecord):
    """Credential exchange anoncreds detail record."""

    class Meta:
        """V20CredExRecordAnonCreds metadata."""

        schema_class = "V20CredExRecordAnonCredsSchema"

    RECORD_ID_NAME = "cred_ex_anoncreds_id"
    RECORD_TYPE = "anoncreds_cred_ex_v20"
    TAG_NAMES = {"~cred_ex_id"} if UNENCRYPTED_TAGS else {"cred_ex_id"}
    RECORD_TOPIC = "issue_credential_v2_0_anoncreds"

    def __init__(
        self,
        cred_ex_anoncreds_id: Optional[str] = None,
        *,
        cred_ex_id: Optional[str] = None,
        cred_id_stored: Optional[str] = None,
        cred_request_metadata: Optional[Mapping] = None,
        rev_reg_id: Optional[str] = None,
        cred_rev_id: Optional[str] = None,
        **kwargs,
    ):
        """Initialize anoncreds credential exchange record details."""
        super().__init__(cred_ex_anoncreds_id, **kwargs)

        self.cred_ex_id = cred_ex_id
        self.cred_id_stored = cred_id_stored
        self.cred_request_metadata = cred_request_metadata
        self.rev_reg_id = rev_reg_id
        self.cred_rev_id = cred_rev_id

    @property
    def cred_ex_anoncreds_id(self) -> str:
        """Accessor for the ID associated with this exchange."""
        return self._id

    @property
    def record_value(self) -> dict:
        """Accessor for the JSON record value generated for this credential exchange."""
        return {
            prop: getattr(self, prop)
            for prop in (
                "cred_id_stored",
                "cred_request_metadata",
                "rev_reg_id",
                "cred_rev_id",
            )
        }

    @classmethod
    async def query_by_cred_ex_id(
        cls,
        session: ProfileSession,
        cred_ex_id: str,
    ) -> Sequence["V20CredExRecordAnonCreds"]:
        """Retrieve credential exchange anoncreds detail record(s) by its cred ex id."""
        return await cls.query(
            session=session,
            tag_filter={"cred_ex_id": cred_ex_id},
        )

    def __eq__(self, other: Any) -> bool:
        """Comparison between records."""
        return super().__eq__(other)


class V20CredExRecordAnonCredsSchema(BaseRecordSchema):
    """Credential exchange anoncreds detail record detail schema."""

    class Meta:
        """Credential exchange anoncreds detail record schema metadata."""

        model_class = V20CredExRecordAnonCreds
        unknown = EXCLUDE

    cred_ex_anoncreds_id = fields.Str(
        required=False,
        metadata={"description": "Record identifier", "example": UUID4_EXAMPLE},
    )
    cred_ex_id = fields.Str(
        required=False,
        metadata={
            "description": "Corresponding v2.0 credential exchange record identifier",
            "example": UUID4_EXAMPLE,
        },
    )
    cred_id_stored = fields.Str(
        required=False,
        metadata={
            "description": "Credential identifier stored in wallet",
            "example": UUID4_EXAMPLE,
        },
    )
    cred_request_metadata = fields.Dict(
        required=False,
        metadata={"description": "Credential request metadata for anoncreds holder"},
    )
    rev_reg_id = fields.Str(
        required=False,
        metadata={
            "description": "Revocation registry identifier",
            "example": ANONCREDS_REV_REG_ID_EXAMPLE,
        },
    )
    cred_rev_id = fields.Str(
        required=False,
        metadata={
            "description": (
                "Credential revocation identifier within revocation registry"
            ),
            "example": ANONCREDS_CRED_DEF_ID_EXAMPLE,
        },
    )



"""Transaction request handler."""

from .....messaging.base_handler import (
    BaseHandler,
    BaseResponder,
    HandlerException,
    RequestContext,
)
from .....storage.error import StorageError
from ..manager import TransactionManager, TransactionManagerError
from ..messages.transaction_request import TransactionRequest
from ..models.transaction_record import TransactionRecord


class TransactionRequestHandler(BaseHandler):
    """Handler class for transaction request."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Handle transaction request.

        Args:
            context: Request context
            responder: Responder callback
        """

        self._logger.debug("TransactionRequestHandler called with context %s", context)
        assert isinstance(context.message, TransactionRequest)

        if not context.connection_ready:
            raise HandlerException("No connection established")

        mgr = TransactionManager(context.profile)
        try:
            transaction = await mgr.receive_request(
                context.message, context.connection_record.connection_id
            )
        except TransactionManagerError as err:
            self._logger.exception(err)
            return

        # Automatically endorse transaction if flag is set
        if context.settings.get("endorser.auto_endorse"):
            try:
                (
                    transaction,
                    endorsed_transaction_response,
                ) = await mgr.create_endorse_response(
                    transaction=transaction,
                    state=TransactionRecord.STATE_TRANSACTION_ENDORSED,
                )

                await responder.send_reply(
                    endorsed_transaction_response,
                    connection_id=transaction.connection_id,
                )
            except (StorageError, TransactionManagerError) as err:
                self._logger.exception(err)



"""Presentation problem report message handler."""

from .....messaging.base_handler import BaseHandler, HandlerException
from .....messaging.request_context import RequestContext
from .....messaging.responder import BaseResponder
from .....storage.error import StorageError, StorageNotFoundError
from ..manager import PresentationManager
from ..messages.presentation_problem_report import PresentationProblemReport


class PresentationProblemReportHandler(BaseHandler):
    """Message handler class for problem reports."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Message handler logic for problem reports.

        Args:
            context: request context
            responder: responder callback
        """
        self._logger.debug(
            "Present-proof v1.0 problem report handler called with context %s",
            context,
        )
        assert isinstance(context.message, PresentationProblemReport)

        # If connection is present it must be ready for use
        if context.connection_record and not context.connection_ready:
            raise HandlerException(
                "Connection used for presentation problem report not ready"
            )

        presentation_manager = PresentationManager(context.profile)
        try:
            await presentation_manager.receive_problem_report(
                context.message,
                (
                    context.connection_record.connection_id
                    if context.connection_record is not None
                    else None
                ),
            )
        except (StorageError, StorageNotFoundError):
            self._logger.exception(
                "Error processing present-proof v1.0 problem report message"
            )



"""Represents a rotate problem report message."""

from enum import Enum

from marshmallow import EXCLUDE, ValidationError, pre_dump, validates_schema

from .....protocols.problem_report.v1_0.message import ProblemReport, ProblemReportSchema
from ..message_types import PROBLEM_REPORT, PROTOCOL_PACKAGE

HANDLER_CLASS = f"{PROTOCOL_PACKAGE}.handlers.problem_report_handler.ProblemReportHandler"


class ProblemReportReason(Enum):
    """Supported reason codes."""

    UNRESOLVABLE = "e.did.unresolvable"
    UNSUPPORTED_METHOD = "e.did.unsupported_method"
    UNRESOLVABLE_SERVICES = "e.did.unresolvable_services"
    UNRECORDABLE_KEYS = "e.did.unrecordable_keys"


class RotateProblemReport(ProblemReport):
    """Base class representing a Rotate problem report message."""

    class Meta:
        """Rotate problem report metadata."""

        handler_class = HANDLER_CLASS
        message_type = PROBLEM_REPORT
        schema_class = "RotateProblemReportSchema"

    @classmethod
    def for_code(cls, problem_code: ProblemReportReason, did: str, **kwargs):
        """Initialize a ProblemReport message instance for a specific problem code.

        Args:
            problem_code: The standard error identifier
            did: The DID contained in the rotate message the problem report is about
            kwargs: Optional keyword arguments to pass through to the problem report
                constructor
        Returns:
            An instance of RotateProblemReport
        """
        description = {
            ProblemReportReason.UNRESOLVABLE: "Unable to resolve DID",
            ProblemReportReason.UNSUPPORTED_METHOD: "Unsupported DID Method",
            ProblemReportReason.UNRESOLVABLE_SERVICES: "Unable to resolve DIDComm Services",  # noqa: E501
            ProblemReportReason.UNRECORDABLE_KEYS: "Unable to record Keys ro Resolvable DID",  # noqa: E501
        }[problem_code]
        return cls(
            description={
                "en": description,
                "code": problem_code.value,
            },
            problem_items=[
                {"did": did},
            ],
            **kwargs,
        )

    @classmethod
    def unresolvable(cls, did: str, **kwargs):
        """Initialize a ProblemReport message instance for an unresolvable DID.

        Args:
            did: The DID contained in the rotate message the problem report is about
            kwargs: Optional keyword arguments to pass through to the problem report
                constructor
        Returns:
            An instance of RotateProblemReport
        """
        return cls.for_code(ProblemReportReason.UNRESOLVABLE, did, **kwargs)

    @classmethod
    def unsupported_method(cls, did: str, **kwargs):
        """Initialize a ProblemReport message instance for an unsupported DID method.

        Args:
            did: The DID contained in the rotate message the problem report is about
            kwargs: Optional keyword arguments to pass through to the problem report
                constructor
        Returns:
            An instance of RotateProblemReport
        """
        return cls.for_code(ProblemReportReason.UNSUPPORTED_METHOD, did, **kwargs)

    @classmethod
    def unresolvable_services(cls, did: str, **kwargs):
        """Initialize a ProblemReport message instance for unresolvable DIDComm services.

        Args:
            did: The DID contained in the rotate message the problem report is about
            kwargs: Optional keyword arguments to pass through to the problem report
                constructor
        Returns:
            An instance of RotateProblemReport
        """
        return cls.for_code(ProblemReportReason.UNRESOLVABLE_SERVICES, did, **kwargs)

    @classmethod
    def unrecordable_keys(cls, did: str, **kwargs):
        """Initialize a ProblemReport message instance for unrecordable keys for resolvable DID.

        Args:
            did: The DID contained in the rotate message the problem report is about
            kwargs: Optional keyword arguments to pass through to the problem report
                constructor
        Returns:
            An instance of RotateProblemReport
        """  # noqa: E501
        return cls.for_code(ProblemReportReason.UNRECORDABLE_KEYS, did, **kwargs)


class RotateProblemReportSchema(ProblemReportSchema):
    """Schema for RotateProblemReport base class."""

    class Meta:
        """Metadata for Rotate problem report schema."""

        model_class = RotateProblemReport
        unknown = EXCLUDE

    @validates_schema
    def validate_fields(self, data, **kwargs):
        """Validate schema fields.

        Args:
            data: The data to validate
            kwargs: Additional keyword arguments

        Raises:
            ValidationError: if data is invalid

        """
        if "problem_items" not in data or not data["problem_items"]:
            raise ValidationError("Rotate problem report must contain problem_items")

        if "did" not in data["problem_items"][0]:
            raise ValidationError("Rotate problem report problem_items must contain did")

    @pre_dump
    def check_thread_deco(self, obj, **kwargs):
        """Thread decorator, and its thid, are mandatory."""
        if not obj._decorators.to_dict().get("~thread", {}).keys() >= {"thid"}:
            raise ValidationError("Missing required field(s) in thread decorator")
        return obj



from unittest import IsolatedAsyncioTestCase

from acapy_agent.tests import mock

from ...config.base import ConfigError
from ...config.error import ArgsParseError
from ...protocols.coordinate_mediation.mediation_invite_store import MediationInviteRecord
from ...utils.testing import create_test_profile
from .. import provision as test_module


class TestProvision(IsolatedAsyncioTestCase):
    def test_bad_calls(self):
        with self.assertRaises(ArgsParseError):
            test_module.execute([])

        with self.assertRaises(SystemExit):
            test_module.execute(["bad"])

    async def test_provision_ledger_configured(self):
        profile = mock.MagicMock(close=mock.CoroutineMock())
        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                mock.CoroutineMock(
                    return_value=(
                        profile,
                        mock.CoroutineMock(did="public DID", verkey="verkey"),
                    )
                ),
            ),
            mock.patch.object(
                test_module, "ledger_config", mock.CoroutineMock(return_value=True)
            ),
        ):
            await test_module.provision({})

    async def test_provision_config_x(self):
        with mock.patch.object(
            test_module, "wallet_config", mock.CoroutineMock()
        ) as mock_wallet_config:
            mock_wallet_config.side_effect = ConfigError("oops")
            with self.assertRaises(test_module.ProvisionError):
                await test_module.provision({})

    def test_main(self):
        with (
            mock.patch.object(test_module, "__name__", "__main__"),
            mock.patch.object(test_module, "execute", mock.MagicMock()) as mock_execute,
        ):
            test_module.main()
            mock_execute.assert_called_once

    async def test_provision_should_store_provided_mediation_invite(self):
        # given
        mediation_invite = "test-invite"
        test_profile = await create_test_profile()

        with (
            mock.patch.object(test_module.MediationInviteStore, "store") as invite_store,
            mock.patch.object(
                test_module,
                "wallet_config",
                mock.CoroutineMock(return_value=(test_profile, mock.MagicMock())),
            ),
        ):
            # when
            await test_module.provision({"mediation.invite": mediation_invite})

            # then
            invite_store.assert_called_with(
                MediationInviteRecord(mediation_invite, False)
            )



from datetime import datetime, timedelta
from unittest import IsolatedAsyncioTestCase, mock

from ...validation_result import PurposeResult
from ..authentication_proof_purpose import AuthenticationProofPurpose
from ..controller_proof_purpose import ControllerProofPurpose


class TestAuthenticationProofPurpose(IsolatedAsyncioTestCase):
    async def test_properties(self):
        date = datetime.now()
        delta = timedelta(1)
        challenge = "challenge"
        domain = "domain"
        proof_purpose = AuthenticationProofPurpose(
            challenge=challenge, domain=domain, date=date, max_timestamp_delta=delta
        )
        proof_purpose2 = AuthenticationProofPurpose(
            challenge=challenge, domain=domain, date=date, max_timestamp_delta=delta
        )

        assert proof_purpose.term == "authentication"
        assert proof_purpose.date == date
        assert proof_purpose.max_timestamp_delta == delta
        assert proof_purpose.challenge == challenge
        assert proof_purpose.domain == domain

        assert proof_purpose2 == proof_purpose
        assert proof_purpose != 10

    async def test_validate(self):
        proof_purpose = AuthenticationProofPurpose(
            challenge="8378c56e-4926-4a54-9587-0f2ef564619a", domain="example.com"
        )

        with mock.patch.object(ControllerProofPurpose, "validate") as validate_mock:
            validate_mock.return_value = PurposeResult(valid=True)
            proof = {
                "challenge": "8378c56e-4926-4a54-9587-0f2ef564619a",
                "domain": "example.com",
            }
            document = mock.MagicMock()
            suite = mock.MagicMock()
            verification_method = {"controller": "controller"}
            document_loader = mock.MagicMock()

            result = proof_purpose.validate(
                proof=proof,
                document=document,
                suite=suite,
                verification_method=verification_method,
                document_loader=document_loader,
            )
            validate_mock.assert_called_once_with(
                proof=proof,
                document=document,
                suite=suite,
                verification_method=verification_method,
                document_loader=document_loader,
            )
            assert result.valid

    async def test_validate_x_challenge_does_not_match(self):
        proof_purpose = AuthenticationProofPurpose(
            challenge="8378c56e-4926-4a54-9587-0f2ef564619a", domain="example.com"
        )

        with mock.patch.object(ControllerProofPurpose, "validate") as validate_mock:
            validate_mock.return_value = PurposeResult(valid=True)

            result = proof_purpose.validate(
                proof={
                    "challenge": "another8378c56e-4926-4a54-9587-0f2ef564619a",
                    "domain": "example.com",
                },
                document=mock.MagicMock(),
                suite=mock.MagicMock(),
                verification_method=mock.MagicMock(),
                document_loader=mock.MagicMock(),
            )
            assert not result.valid
            assert "The challenge is not as expected" in str(result.error)

    async def test_validate_x_domain_does_not_match(self):
        proof_purpose = AuthenticationProofPurpose(
            challenge="8378c56e-4926-4a54-9587-0f2ef564619a", domain="example.com"
        )

        with mock.patch.object(ControllerProofPurpose, "validate") as validate_mock:
            validate_mock.return_value = PurposeResult(valid=True)

            result = proof_purpose.validate(
                proof={
                    "challenge": "8378c56e-4926-4a54-9587-0f2ef564619a",
                    "domain": "anotherexample.com",
                },
                document=mock.MagicMock(),
                suite=mock.MagicMock(),
                verification_method=mock.MagicMock(),
                document_loader=mock.MagicMock(),
            )
            assert not result.valid
            assert "The domain is not as expected" in str(result.error)

    async def test_update(self):
        proof_purpose = AuthenticationProofPurpose(
            challenge="8378c56e-4926-4a54-9587-0f2ef564619a", domain="example.com"
        )

        proof = {}
        ret_proof = proof_purpose.update(proof)

        assert proof == ret_proof
        assert proof.get("challenge") == "8378c56e-4926-4a54-9587-0f2ef564619a"
        assert proof.get("domain") == "example.com"



"""Presentation request message handler."""

from .....core.oob_processor import OobMessageProcessor
from .....indy.holder import IndyHolder, IndyHolderError
from .....indy.models.xform import indy_proof_req_preview2indy_requested_creds
from .....ledger.error import LedgerError
from .....messaging.base_handler import BaseHandler, HandlerException
from .....messaging.models.base import BaseModelError
from .....messaging.request_context import RequestContext
from .....messaging.responder import BaseResponder
from .....storage.error import StorageError, StorageNotFoundError
from .....utils.tracing import get_timer, trace_event
from .....wallet.error import WalletNotFoundError
from .. import problem_report_for_record
from ..manager import PresentationManager
from ..messages.presentation_problem_report import ProblemReportReason
from ..messages.presentation_request import PresentationRequest
from ..models.presentation_exchange import V10PresentationExchange


class PresentationRequestHandler(BaseHandler):
    """Message handler class for Aries#0037 v1.0 presentation requests."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Message handler logic for Aries#0037 v1.0 presentation requests.

        Args:
            context: request context
            responder: responder callback

        """
        r_time = get_timer()
        profile = context.profile

        self._logger.debug("PresentationRequestHandler called with context %s", context)
        assert isinstance(context.message, PresentationRequest)
        self._logger.debug(
            "Received presentation request message: %s",
            context.message.serialize(as_string=True),
        )

        # If connection is present it must be ready for use
        if context.connection_record and not context.connection_ready:
            raise HandlerException("Connection used for presentation request not ready")

        # Find associated oob record
        oob_processor = context.inject(OobMessageProcessor)
        oob_record = await oob_processor.find_oob_record_for_inbound_message(context)

        # Either connection or oob context must be present
        if not context.connection_record and not oob_record:
            raise HandlerException(
                "No connection or associated connectionless exchange found for"
                " presentation request"
            )

        connection_id = (
            context.connection_record.connection_id if context.connection_record else None
        )

        presentation_manager = PresentationManager(profile)

        indy_proof_request = context.message.indy_proof_request(0)

        # Get presentation exchange record (holder initiated via proposal)
        # or create it (verifier sent request first)
        try:
            async with profile.session() as session:
                (
                    presentation_exchange_record
                ) = await V10PresentationExchange.retrieve_by_tag_filter(
                    session,
                    {"thread_id": context.message._thread_id},
                    {
                        "role": V10PresentationExchange.ROLE_PROVER,
                        "connection_id": connection_id,
                    },
                )  # holder initiated via proposal
                presentation_exchange_record.presentation_request = indy_proof_request
                presentation_exchange_record.presentation_request_dict = (
                    context.message.serialize()
                )
        except StorageNotFoundError:  # verifier sent this request free of any proposal
            presentation_exchange_record = V10PresentationExchange(
                connection_id=connection_id,
                thread_id=context.message._thread_id,
                initiator=V10PresentationExchange.INITIATOR_EXTERNAL,
                role=V10PresentationExchange.ROLE_PROVER,
                presentation_request=indy_proof_request,
                presentation_request_dict=context.message.serialize(),
                auto_present=context.settings.get(
                    "debug.auto_respond_presentation_request"
                ),
                trace=(context.message._trace is not None),
                auto_remove=not profile.settings.get("preserve_exchange_records"),
            )

        presentation_exchange_record = await presentation_manager.receive_request(
            presentation_exchange_record
        )  # mgr only saves record: on exception, saving state null is hopeless

        r_time = trace_event(
            context.settings,
            context.message,
            outcome="PresentationRequestHandler.handle.END",
            perf_counter=r_time,
        )

        # If auto_present is enabled, respond immediately with presentation
        if presentation_exchange_record.auto_present:
            presentation_preview = None
            if presentation_exchange_record.presentation_proposal_dict:
                exchange_pres_proposal = (
                    presentation_exchange_record.presentation_proposal_dict
                )
                presentation_preview = exchange_pres_proposal.presentation_proposal

            try:
                req_creds = await indy_proof_req_preview2indy_requested_creds(
                    indy_proof_request,
                    presentation_preview,
                    holder=context.inject(IndyHolder),
                )
            except ValueError as err:
                self._logger.warning(f"{err}")
                return  # not a protocol error: prover could still build proof manually

            presentation_message = None
            try:
                (
                    presentation_exchange_record,
                    presentation_message,
                ) = await presentation_manager.create_presentation(
                    presentation_exchange_record=presentation_exchange_record,
                    requested_credentials=req_creds,
                    comment="auto-presented for proof request nonce={}".format(
                        indy_proof_request["nonce"]
                    ),
                )
                await responder.send_reply(presentation_message)
            except (
                BaseModelError,
                IndyHolderError,
                LedgerError,
                StorageError,
                WalletNotFoundError,
            ) as err:
                self._logger.exception(err)
                if presentation_exchange_record:
                    async with profile.session() as session:
                        await presentation_exchange_record.save_error_state(
                            session,
                            reason=err.roll_up,  # us: be specific
                        )
                    await responder.send_reply(
                        problem_report_for_record(
                            presentation_exchange_record,
                            ProblemReportReason.ABANDONED.value,  # them: be vague
                        )
                    )

            trace_event(
                context.settings,
                presentation_message,
                outcome="PresentationRequestHandler.handle.PRESENT",
                perf_counter=r_time,
            )



"""Standard Injector implementation."""

from typing import Dict, Mapping, Optional, Type

from .base import BaseInjector, BaseProvider, InjectionError, InjectType
from .provider import CachedProvider, InstanceProvider
from .settings import Settings


class Injector(BaseInjector):
    """Injector implementation with static and dynamic bindings."""

    def __init__(
        self,
        settings: Optional[Mapping[str, object]] = None,
        *,
        enforce_typing: bool = True,
    ):
        """Initialize an `Injector`."""
        self.enforce_typing = enforce_typing
        self._providers: Dict[Type, BaseProvider] = {}
        self._settings = Settings(settings)

    @property
    def settings(self) -> Settings:
        """Accessor for scope-specific settings."""
        return self._settings

    @settings.setter
    def settings(self, settings: Settings):
        """Setter for scope-specific settings."""
        self._settings = settings

    def bind_instance(self, base_cls: Type[InjectType], instance: InjectType):
        """Add a static instance as a class binding."""
        self._providers[base_cls] = InstanceProvider(instance)

    def bind_provider(
        self, base_cls: Type[InjectType], provider: BaseProvider, *, cache: bool = False
    ):
        """Add a dynamic instance resolver as a class binding."""
        if not provider:
            raise ValueError("Class provider binding must be non-empty")
        if cache and not isinstance(provider, CachedProvider):
            provider = CachedProvider(provider)
        self._providers[base_cls] = provider

    def soft_bind_instance(self, base_cls: Type[InjectType], instance: InjectType):
        """Add a static instance as a soft class binding.

        The binding occurs only if a provider for the same type does not already exist.
        """
        if not self.get_provider(base_cls):
            self.bind_instance(base_cls, instance)

    def soft_bind_provider(
        self, base_cls: Type[InjectType], provider: BaseProvider, *, cache: bool = False
    ):
        """Add a dynamic instance resolver as a soft class binding.

        The binding occurs only if a provider for the same type does not already exist.
        """
        if not self.get_provider(base_cls):
            self.bind_provider(base_cls, provider, cache=cache)

    def clear_binding(self, base_cls: Type[InjectType]):
        """Remove a previously-added binding."""
        if base_cls in self._providers:
            del self._providers[base_cls]

    def get_provider(self, base_cls: Type[InjectType]):
        """Find the provider associated with a class binding."""
        return self._providers.get(base_cls)

    def inject_or(
        self,
        base_cls: Type[InjectType],
        settings: Optional[Mapping[str, object]] = None,
        default: Optional[InjectType] = None,
    ) -> Optional[InjectType]:
        """Get the provided instance of a given class identifier or default if not found.

        Args:
            base_cls: The base class to retrieve an instance of
            settings: An optional dict providing configuration to the provider
            default: default return value if no instance is found

        Returns:
            An instance of the base class, or None

        """
        if not base_cls:
            raise InjectionError("No base class provided for lookup")
        provider = self._providers.get(base_cls)
        if settings:
            ext_settings = self.settings.extend(settings)
        else:
            ext_settings = self.settings
        if provider:
            result = provider.provide(ext_settings, self)
        else:
            result = None

        if result and not isinstance(result, base_cls) and self.enforce_typing:
            raise InjectionError(
                "Provided instance does not implement the base class: {}".format(
                    base_cls.__name__
                )
            )

        return result if result is not None else default

    def inject(
        self,
        base_cls: Type[InjectType],
        settings: Optional[Mapping[str, object]] = None,
    ) -> InjectType:
        """Get the provided instance of a given class identifier.

        Args:
            base_cls (Type[InjectType]): The base class to retrieve an instance of.
            settings (Optional[Mapping[str, object]]): An optional dictionary providing
                configuration to the provider.

        Returns:
            InjectType: An instance of the base class, or None.

        Raises:
            InjectionError: If no instance is provided for the given class identifier.

        """
        result = self.inject_or(base_cls, settings)
        if result is None:
            raise InjectionError(
                "No instance provided for class: {}".format(base_cls.__name__)
            )
        return result

    def copy(self) -> BaseInjector:
        """Produce a copy of the injector instance."""
        result = Injector(self.settings)
        result.enforce_typing = self.enforce_typing
        result._providers = self._providers.copy()
        return result

    def __repr__(self) -> str:
        """Provide a human readable representation of this object."""
        return f"<{self.__class__.__name__}>"



from unittest import TestCase

from ...wallet.key_type import P256
from ...wallet.util import b58_to_bytes
from ..did_key import DID_KEY_RESOLVERS, DIDKey
from .test_dids import DID_P256_zDnaerDaTF5BXEavCrfRZEk316dpbLsfPDZ3WJ5hRTPFU2169

TEST_P256_BASE58_KEY = "23FF9c3MrW7NkEW6uNDvdSKQMJ4YFTBXNMEPytZfYeE33"
TEST_P256_FINGERPRINT = "zDnaerDaTF5BXEavCrfRZEk316dpbLsfPDZ3WJ5hRTPFU2169"
TEST_P256_DID = f"did:key:{TEST_P256_FINGERPRINT}"
TEST_P256_KEY_ID = f"{TEST_P256_DID}#{TEST_P256_FINGERPRINT}"
TEST_P256_PREFIX_BYTES = b"".join([b"\x80\x24", b58_to_bytes(TEST_P256_BASE58_KEY)])


class TestDIDKey(TestCase):
    def test_p256_from_public_key(self):
        key_bytes = b58_to_bytes(TEST_P256_BASE58_KEY)
        did_key = DIDKey.from_public_key(key_bytes, P256)

        assert did_key.did == TEST_P256_DID

    def test_p256_from_public_key_b58(self):
        did_key = DIDKey.from_public_key_b58(TEST_P256_BASE58_KEY, P256)

        assert did_key.did == TEST_P256_DID

    def test_p256_from_fingerprint(self):
        did_key = DIDKey.from_fingerprint(TEST_P256_FINGERPRINT)

        assert did_key.did == TEST_P256_DID
        assert did_key.public_key_b58 == TEST_P256_BASE58_KEY

    def test_p256_from_did(self):
        did_key = DIDKey.from_did(TEST_P256_DID)

        assert did_key.public_key_b58 == TEST_P256_BASE58_KEY

    def test_p256_properties(self):
        did_key = DIDKey.from_did(TEST_P256_DID)

        assert did_key.fingerprint == TEST_P256_FINGERPRINT
        assert did_key.did == TEST_P256_DID
        assert did_key.public_key_b58 == TEST_P256_BASE58_KEY
        assert did_key.public_key == b58_to_bytes(TEST_P256_BASE58_KEY)
        assert did_key.key_type == P256
        assert did_key.key_id == TEST_P256_KEY_ID
        assert did_key.prefixed_public_key == TEST_P256_PREFIX_BYTES

    def test_p256_diddoc(self):
        did_key = DIDKey.from_did(TEST_P256_DID)

        resolver = DID_KEY_RESOLVERS[P256]

        assert resolver(did_key) == did_key.did_doc

    def test_p256_resolver(self):
        did_key = DIDKey.from_did(TEST_P256_DID)
        resolver = DID_KEY_RESOLVERS[P256]
        did_doc = resolver(did_key)

        assert did_doc == DID_P256_zDnaerDaTF5BXEavCrfRZEk316dpbLsfPDZ3WJ5hRTPFU2169



from unittest import IsolatedAsyncioTestCase

from ....admin.request_context import AdminRequestContext
from ....connections.models.conn_record import ConnRecord
from ....indy.issuer import IndyIssuer
from ....ledger.base import BaseLedger
from ....ledger.multiple_ledger.ledger_requests_executor import IndyLedgerRequestsExecutor
from ....multitenant.base import BaseMultitenantManager
from ....multitenant.manager import MultitenantManager
from ....storage.base import BaseStorage
from ....tests import mock
from ....utils.testing import create_test_profile
from .. import routes as test_module

SCHEMA_ID = "WgWxqztrNooG92RXvxSTWv:2:schema_name:1.0"


class TestSchemaRoutes(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.session_inject = {}
        self.profile = await create_test_profile(
            settings={
                "admin.admin_api_key": "secret-key",
            }
        )
        self.profile_injector = self.profile.context.injector
        self.ledger = mock.create_autospec(BaseLedger)
        self.ledger.__aenter__ = mock.CoroutineMock(return_value=self.ledger)
        self.ledger.create_and_send_schema = mock.CoroutineMock(
            return_value=(SCHEMA_ID, {"schema": "def", "signed_txn": "..."})
        )
        self.ledger.get_schema = mock.CoroutineMock(
            return_value={"schema": "def", "signed_txn": "..."}
        )
        self.profile_injector.bind_instance(BaseLedger, self.ledger)

        self.issuer = mock.create_autospec(IndyIssuer)
        self.profile_injector.bind_instance(IndyIssuer, self.issuer)

        self.storage = mock.create_autospec(BaseStorage)
        self.storage.find_all_records = mock.CoroutineMock(
            return_value=[mock.MagicMock(value=SCHEMA_ID)]
        )
        self.session_inject[BaseStorage] = self.storage
        self.context = AdminRequestContext.test_context(
            self.session_inject, profile=self.profile
        )
        self.request_dict = {
            "context": self.context,
            "outbound_message_router": mock.CoroutineMock(),
        }
        self.request = mock.MagicMock(
            app={},
            match_info={},
            query={},
            __getitem__=lambda _, k: self.request_dict[k],
            headers={"x-api-key": "secret-key"},
        )

    async def test_send_schema(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "schema_name": "schema_name",
                "schema_version": "1.0",
                "attributes": ["table", "drink", "colour"],
            }
        )

        self.request.query = {"create_transaction_for_endorser": "false"}

        with mock.patch.object(test_module.web, "json_response") as mock_response:
            result = await test_module.schemas_send_schema(self.request)
            assert result == mock_response.return_value
            mock_response.assert_called_once_with(
                {
                    "sent": {
                        "schema_id": SCHEMA_ID,
                        "schema": {
                            "schema": "def",
                            "signed_txn": "...",
                        },
                    },
                    "schema_id": SCHEMA_ID,
                    "schema": {
                        "schema": "def",
                        "signed_txn": "...",
                    },
                }
            )

    async def test_send_schema_create_transaction_for_endorser(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "schema_name": "schema_name",
                "schema_version": "1.0",
                "attributes": ["table", "drink", "colour"],
            }
        )

        self.request.query = {
            "create_transaction_for_endorser": "true",
            "conn_id": "dummy",
        }

        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                test_module, "TransactionManager", mock.MagicMock()
            ) as mock_txn_mgr,
            mock.patch.object(
                test_module.web, "json_response", mock.MagicMock()
            ) as mock_response,
        ):
            mock_txn_mgr.return_value = mock.MagicMock(
                create_record=mock.CoroutineMock(
                    return_value=mock.MagicMock(
                        serialize=mock.MagicMock(return_value={"...": "..."})
                    )
                )
            )
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "endorser_did": ("did"),
                        "endorser_name": ("name"),
                    }
                )
            )
            result = await test_module.schemas_send_schema(self.request)
            assert result == mock_response.return_value
            mock_response.assert_called_once_with(
                {
                    "sent": {
                        "schema_id": SCHEMA_ID,
                        "schema": {
                            "schema": "def",
                            "signed_txn": "...",
                        },
                    },
                    "txn": {"...": "..."},
                }
            )

    async def test_send_schema_create_transaction_for_endorser_storage_x(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "schema_name": "schema_name",
                "schema_version": "1.0",
                "attributes": ["table", "drink", "colour"],
            }
        )

        self.request.query = {
            "create_transaction_for_endorser": "true",
            "conn_id": "dummy",
        }

        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                test_module, "TransactionManager", mock.MagicMock()
            ) as mock_txn_mgr,
        ):
            mock_txn_mgr.return_value = mock.MagicMock(
                create_record=mock.CoroutineMock(side_effect=test_module.StorageError())
            )
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "endorser_did": ("did"),
                        "endorser_name": ("name"),
                    }
                )
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.schemas_send_schema(self.request)

    async def test_send_schema_create_transaction_for_endorser_not_found_x(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "schema_name": "schema_name",
                "schema_version": "1.0",
                "attributes": ["table", "drink", "colour"],
            }
        )

        self.request.query = {
            "create_transaction_for_endorser": "true",
            "conn_id": "dummy",
        }

        with mock.patch.object(
            ConnRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_conn_rec_retrieve:
            mock_conn_rec_retrieve.side_effect = test_module.StorageNotFoundError()

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.schemas_send_schema(self.request)

    async def test_send_schema_create_transaction_for_endorser_base_model_x(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "schema_name": "schema_name",
                "schema_version": "1.0",
                "attributes": ["table", "drink", "colour"],
            }
        )

        self.request.query = {
            "create_transaction_for_endorser": "true",
            "conn_id": "dummy",
        }

        with mock.patch.object(
            ConnRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_conn_rec_retrieve:
            mock_conn_rec_retrieve.side_effect = test_module.BaseModelError()

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.schemas_send_schema(self.request)

    async def test_send_schema_create_transaction_for_endorser_no_endorser_info_x(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "schema_name": "schema_name",
                "schema_version": "1.0",
                "attributes": ["table", "drink", "colour"],
            }
        )

        self.request.query = {
            "create_transaction_for_endorser": "true",
            "conn_id": "dummy",
        }

        with mock.patch.object(
            ConnRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_conn_rec_retrieve:
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(return_value=None)
            )
            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.schemas_send_schema(self.request)

    async def test_send_schema_create_transaction_for_endorser_no_endorser_did_x(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "schema_name": "schema_name",
                "schema_version": "1.0",
                "attributes": ["table", "drink", "colour"],
            }
        )

        self.request.query = {
            "create_transaction_for_endorser": "true",
            "conn_id": "dummy",
        }

        with mock.patch.object(
            ConnRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_conn_rec_retrieve:
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "endorser_name": ("name"),
                    }
                )
            )
            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.schemas_send_schema(self.request)

    async def test_send_schema_no_ledger(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "schema_name": "schema_name",
                "schema_version": "1.0",
                "attributes": ["table", "drink", "colour"],
            }
        )

        self.context.injector.clear_binding(BaseLedger)
        with self.assertRaises(test_module.web.HTTPForbidden):
            await test_module.schemas_send_schema(self.request)

    async def test_send_schema_x_ledger(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "schema_name": "schema_name",
                "schema_version": "1.0",
                "attributes": ["table", "drink", "colour"],
            }
        )
        self.request.query = {"create_transaction_for_endorser": "false"}
        self.ledger.create_and_send_schema = mock.CoroutineMock(
            side_effect=test_module.LedgerError("Down for routine maintenance")
        )

        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.schemas_send_schema(self.request)

    async def test_created(self):
        self.request.match_info = {"schema_id": SCHEMA_ID}

        with mock.patch.object(test_module.web, "json_response") as mock_response:
            result = await test_module.schemas_created(self.request)
            assert result == mock_response.return_value
            mock_response.assert_called_once_with({"schema_ids": [SCHEMA_ID]})

    async def test_get_schema(self):
        self.ledger.get_schema = mock.CoroutineMock(
            side_effect=[{"schema": "def", "signed_txn": "..."}, None]
        )
        mock_executor = mock.MagicMock(IndyLedgerRequestsExecutor, autospec=True)
        mock_executor.get_ledger_for_identifier = mock.CoroutineMock(
            return_value=("test_ledger_id", self.ledger)
        )
        self.profile_injector.bind_instance(
            IndyLedgerRequestsExecutor,
            mock_executor,
        )
        self.request.match_info = {"schema_id": SCHEMA_ID}
        with mock.patch.object(test_module.web, "json_response") as mock_response:
            result = await test_module.schemas_get_schema(self.request)
            assert result == mock_response.return_value
            mock_response.assert_called_once_with(
                {
                    "ledger_id": "test_ledger_id",
                    "schema": {"schema": "def", "signed_txn": "..."},
                }
            )

        # test schema not found
        with self.assertRaises(test_module.web.HTTPNotFound):
            await test_module.schemas_get_schema(self.request)

    async def test_get_schema_multitenant(self):
        self.profile_injector.bind_instance(
            BaseMultitenantManager,
            mock.MagicMock(MultitenantManager, autospec=True),
        )
        self.request.match_info = {"schema_id": SCHEMA_ID}
        with (
            mock.patch.object(
                IndyLedgerRequestsExecutor,
                "get_ledger_for_identifier",
                mock.CoroutineMock(return_value=("test_ledger_id", self.ledger)),
            ),
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            result = await test_module.schemas_get_schema(self.request)
            assert result == mock_response.return_value
            mock_response.assert_called_once_with(
                {
                    "ledger_id": "test_ledger_id",
                    "schema": {"schema": "def", "signed_txn": "..."},
                }
            )

    async def test_get_schema_on_seq_no(self):
        mock_executor = mock.MagicMock(IndyLedgerRequestsExecutor, autospec=True)
        mock_executor.get_ledger_for_identifier = mock.CoroutineMock(
            return_value=(None, self.ledger)
        )
        self.profile_injector.bind_instance(
            IndyLedgerRequestsExecutor,
            mock_executor,
        )
        self.request.match_info = {"schema_id": "12345"}
        with mock.patch.object(test_module.web, "json_response") as mock_response:
            result = await test_module.schemas_get_schema(self.request)
            assert result == mock_response.return_value
            mock_response.assert_called_once_with(
                {"schema": {"schema": "def", "signed_txn": "..."}}
            )

    async def test_get_schema_no_ledger(self):
        mock_executor = mock.MagicMock(IndyLedgerRequestsExecutor, autospec=True)
        mock_executor.get_ledger_for_identifier = mock.CoroutineMock(
            return_value=(None, None)
        )
        self.profile_injector.bind_instance(
            IndyLedgerRequestsExecutor,
            mock_executor,
        )
        self.request.match_info = {"schema_id": SCHEMA_ID}
        self.ledger.get_schema = mock.CoroutineMock(
            side_effect=test_module.LedgerError("Down for routine maintenance")
        )

        self.context.injector.clear_binding(BaseLedger)
        with self.assertRaises(test_module.web.HTTPForbidden):
            await test_module.schemas_get_schema(self.request)

    async def test_get_schema_x_ledger(self):
        mock_executor = mock.MagicMock(IndyLedgerRequestsExecutor, autospec=True)
        mock_executor.get_ledger_for_identifier = mock.CoroutineMock(
            return_value=(None, self.ledger)
        )
        self.profile_injector.bind_instance(
            IndyLedgerRequestsExecutor,
            mock_executor,
        )
        self.request.match_info = {"schema_id": SCHEMA_ID}
        self.ledger.get_schema = mock.CoroutineMock(
            side_effect=test_module.LedgerError("Down for routine maintenance")
        )

        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.schemas_get_schema(self.request)

    async def test_register(self):
        mock_app = mock.MagicMock()
        mock_app.add_routes = mock.MagicMock()

        await test_module.register(mock_app)
        mock_app.add_routes.assert_called_once()

    async def test_post_process_routes(self):
        mock_app = mock.MagicMock(_state={"swagger_dict": {}})
        test_module.post_process_routes(mock_app)
        assert "tags" in mock_app._state["swagger_dict"]



from unittest import IsolatedAsyncioTestCase, TestCase, mock

from .....didcomm_prefix import DIDCommPrefix
from ...message_types import CANCEL_TRANSACTION
from ..cancel_transaction import CancelTransaction


class TestConfig:
    test_thread_id = "3fa85f64-5717-4562-b3fc-2c963f66afa6"
    test_state = "cancelled"


class TestCancelTransaction(TestCase, TestConfig):
    def setUp(self):
        self.cancel_transaction = CancelTransaction(
            thread_id=self.test_thread_id, state=self.test_state
        )

    def test_init(self):
        """Test initialization."""
        assert self.cancel_transaction.thread_id == self.test_thread_id
        assert self.cancel_transaction.state == self.test_state

    def test_type(self):
        """Test type."""
        assert self.cancel_transaction._type == DIDCommPrefix.qualify_current(
            CANCEL_TRANSACTION
        )

    @mock.patch(
        "acapy_agent.protocols.endorse_transaction.v1_0.messages."
        "cancel_transaction.CancelTransactionSchema.load"
    )
    def test_deserialize(self, mock_cancel_transaction_schema_load):
        """
        Test deserialization.
        """
        obj = self.cancel_transaction

        cancel_transaction = CancelTransaction.deserialize(obj)
        mock_cancel_transaction_schema_load.assert_called_once_with(obj)

        assert cancel_transaction is mock_cancel_transaction_schema_load.return_value

    @mock.patch(
        "acapy_agent.protocols.endorse_transaction.v1_0.messages."
        "cancel_transaction.CancelTransactionSchema.dump"
    )
    def test_serialize(self, mock_cancel_transaction_schema_dump):
        """
        Test serialization.
        """
        cancel_transaction_dict = self.cancel_transaction.serialize()
        mock_cancel_transaction_schema_dump.assert_called_once_with(
            self.cancel_transaction
        )

        assert cancel_transaction_dict is mock_cancel_transaction_schema_dump.return_value


class TestCancelTransactionSchema(IsolatedAsyncioTestCase, TestConfig):
    """Test cancel transaction schema."""

    async def test_make_model(self):
        cancel_transaction = CancelTransaction(
            thread_id=self.test_thread_id, state=self.test_state
        )
        data = cancel_transaction.serialize()
        model_instance = CancelTransaction.deserialize(data)
        assert type(model_instance) is type(cancel_transaction)



"""Endorse Transaction handling admin routes."""

import logging

from aiohttp import web
from aiohttp_apispec import (
    docs,
    match_info_schema,
    querystring_schema,
    request_schema,
    response_schema,
)
from marshmallow import fields, validate

from ....admin.decorators.auth import tenant_authentication
from ....admin.request_context import AdminRequestContext
from ....connections.models.conn_record import ConnRecord
from ....core.event_bus import Event, EventBus
from ....core.profile import Profile
from ....core.util import SHUTDOWN_EVENT_PATTERN, STARTUP_EVENT_PATTERN
from ....indy.issuer import IndyIssuerError
from ....ledger.error import LedgerError
from ....messaging.models.base import BaseModelError
from ....messaging.models.openapi import OpenAPISchema
from ....messaging.valid import UUID4_EXAMPLE
from ....storage.error import StorageError, StorageNotFoundError
from ....utils.endorsement_setup import attempt_auto_author_with_endorser_setup
from .manager import TransactionManager, TransactionManagerError
from .models.transaction_record import TransactionRecord, TransactionRecordSchema
from .transaction_jobs import TransactionJob

LOGGER = logging.getLogger(__name__)


class TransactionListSchema(OpenAPISchema):
    """Result schema for transaction list."""

    results = fields.List(
        fields.Nested(TransactionRecordSchema()),
        metadata={"description": "List of transaction records"},
    )


class TransactionsListQueryStringSchema(OpenAPISchema):
    """Parameters and validators for transactions list request query string."""


class TranIdMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking transaction id."""

    tran_id = fields.Str(
        required=True,
        metadata={"description": "Transaction identifier", "example": UUID4_EXAMPLE},
    )


class EndorserDIDInfoSchema(OpenAPISchema):
    """Path parameters and validators for request Endorser DID."""

    endorser_did = fields.Str(required=False, metadata={"description": "Endorser DID"})


class AssignTransactionJobsSchema(OpenAPISchema):
    """Assign transaction related jobs to connection record."""

    transaction_my_job = fields.Str(
        required=False,
        validate=validate.OneOf(
            [r.name for r in TransactionJob if isinstance(r.value[0], int)] + ["reset"]
        ),
        metadata={"description": "Transaction related jobs"},
    )


class TransactionJobsSchema(OpenAPISchema):
    """Transaction jobs metadata on connection record."""

    transaction_my_job = fields.Str(
        required=False,
        validate=validate.OneOf(
            [r.name for r in TransactionJob if isinstance(r.value[0], int)] + ["reset"]
        ),
        metadata={"description": "My transaction related job"},
    )
    transaction_their_job = fields.Str(
        required=False,
        validate=validate.OneOf(
            [r.name for r in TransactionJob if isinstance(r.value[0], int)] + ["reset"]
        ),
        metadata={"description": "Their transaction related job"},
    )


class TransactionConnIdMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking connection id."""

    conn_id = fields.Str(
        required=True,
        metadata={"description": "Connection identifier", "example": UUID4_EXAMPLE},
    )


class DateSchema(OpenAPISchema):
    """Sets Expiry date, till when the transaction should be endorsed."""

    expires_time = fields.DateTime(
        required=True,
        metadata={"description": "Expiry Date", "example": "2021-03-29T05:22:19Z"},
    )


class EndorserInfoSchema(OpenAPISchema):
    """Class for user to input the DID associated with the requested endorser."""

    endorser_did = fields.Str(required=True, metadata={"description": "Endorser DID"})

    endorser_name = fields.Str(required=False, metadata={"description": "Endorser Name"})


@docs(
    tags=["endorse-transaction"],
    summary="Query transactions",
)
@querystring_schema(TransactionsListQueryStringSchema())
@response_schema(TransactionListSchema(), 200)
@tenant_authentication
async def transactions_list(request: web.BaseRequest):
    """Request handler for searching transaction records.

    Args:
        request: aiohttp request object
    Returns:
        The transaction list response
    """

    context: AdminRequestContext = request["context"]

    tag_filter = {}
    post_filter = {}

    try:
        async with context.profile.session() as session:
            records = await TransactionRecord.query(
                session, tag_filter, post_filter_positive=post_filter, alt=True
            )
        results = [record.serialize() for record in records]
    except (StorageError, BaseModelError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({"results": results})


@docs(tags=["endorse-transaction"], summary="Fetch a single transaction record")
@match_info_schema(TranIdMatchInfoSchema())
@response_schema(TransactionRecordSchema(), 200)
@tenant_authentication
async def transactions_retrieve(request: web.BaseRequest):
    """Request handler for fetching a single transaction record.

    Args:
        request: aiohttp request object
    Returns:
        The transaction record response
    """

    context: AdminRequestContext = request["context"]
    transaction_id = request.match_info["tran_id"]

    try:
        async with context.profile.session() as session:
            record = await TransactionRecord.retrieve_by_id(session, transaction_id)
        result = record.serialize()
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except BaseModelError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response(result)


# todo - implementing changes for writing final transaction to the ledger
# (For Sign Transaction Protocol)
@docs(
    tags=["endorse-transaction"],
    summary="For author to send a transaction request",
)
@querystring_schema(TranIdMatchInfoSchema())
@request_schema(DateSchema())
@response_schema(TransactionRecordSchema(), 200)
@tenant_authentication
async def transaction_create_request(request: web.BaseRequest):
    """Request handler for creating a new transaction record and request.

    Args:
        request: aiohttp request object
    Returns:
        The transaction record
    """

    context: AdminRequestContext = request["context"]
    outbound_handler = request["outbound_message_router"]
    transaction_id = request.query.get("tran_id")
    endorser_write_txn = False

    body = await request.json()
    expires_time = body.get("expires_time")

    try:
        async with context.profile.session() as session:
            transaction_record = await TransactionRecord.retrieve_by_id(
                session, transaction_id
            )
            connection_record = await ConnRecord.retrieve_by_id(
                session, transaction_record.connection_id
            )
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except BaseModelError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    async with context.profile.session() as session:
        jobs = await connection_record.metadata_get(session, "transaction_jobs")
    if not jobs:
        raise web.HTTPForbidden(
            reason=(
                "The transaction related jobs are not set up in "
                "connection metadata for this connection record"
            )
        )
    if "transaction_my_job" not in jobs.keys():
        raise web.HTTPForbidden(
            reason=(
                'The "transaction_my_job" is not set in "transaction_jobs" '
                "connection metadata for this connection record"
            )
        )
    if "transaction_their_job" not in jobs.keys():
        raise web.HTTPForbidden(
            reason=(
                'Ask the other agent to set up "transaction_my_job" in '
                '"transaction_jobs" in connection metadata for their connection record'
            )
        )
    if jobs["transaction_my_job"] != TransactionJob.TRANSACTION_AUTHOR.name:
        raise web.HTTPForbidden(reason="Only a TRANSACTION_AUTHOR can create a request")

    if jobs["transaction_their_job"] != TransactionJob.TRANSACTION_ENDORSER.name:
        raise web.HTTPForbidden(
            reason="A request can only be created to a TRANSACTION_ENDORSER"
        )

    transaction_mgr = TransactionManager(context.profile)
    try:
        (
            transaction_record,
            transaction_request,
        ) = await transaction_mgr.create_request(
            transaction=transaction_record,
            expires_time=expires_time,
            endorser_write_txn=endorser_write_txn,
        )
    except (StorageError, TransactionManagerError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    await outbound_handler(
        transaction_request, connection_id=connection_record.connection_id
    )

    return web.json_response(transaction_record.serialize())


# todo - implementing changes for writing final transaction to the ledger
# (For Sign Transaction Protocol)
@docs(
    tags=["endorse-transaction"],
    summary="For Endorser to endorse a particular transaction record",
)
@querystring_schema(EndorserDIDInfoSchema())
@match_info_schema(TranIdMatchInfoSchema())
@response_schema(TransactionRecordSchema(), 200)
@tenant_authentication
async def endorse_transaction_response(request: web.BaseRequest):
    """Request handler for creating an endorsed transaction response.

    Args:
        request: aiohttp request object
    Returns:
        The updated transaction record details
    """

    context: AdminRequestContext = request["context"]
    outbound_handler = request["outbound_message_router"]

    transaction_id = request.match_info["tran_id"]
    endorser_did = request.query.get("endorser_did")
    try:
        async with context.profile.session() as session:
            transaction = await TransactionRecord.retrieve_by_id(session, transaction_id)
            connection_record = await ConnRecord.retrieve_by_id(
                session, transaction.connection_id
            )
            # provided DID is validated in the TransactionManager

    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except BaseModelError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    async with context.profile.session() as session:
        jobs = await connection_record.metadata_get(session, "transaction_jobs")
    if not jobs:
        raise web.HTTPForbidden(
            reason=(
                "The transaction related jobs are not set up in "
                "connection metadata for this connection record"
            )
        )
    if jobs["transaction_my_job"] != TransactionJob.TRANSACTION_ENDORSER.name:
        raise web.HTTPForbidden(
            reason="Only a TRANSACTION_ENDORSER can endorse a transaction"
        )

    transaction_mgr = TransactionManager(context.profile)
    try:
        (
            transaction,
            endorsed_transaction_response,
        ) = await transaction_mgr.create_endorse_response(
            transaction=transaction,
            state=TransactionRecord.STATE_TRANSACTION_ENDORSED,
            use_endorser_did=endorser_did,
        )
    except (IndyIssuerError, LedgerError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err
    except (StorageError, TransactionManagerError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    await outbound_handler(
        endorsed_transaction_response, connection_id=transaction.connection_id
    )

    return web.json_response(transaction.serialize())


@docs(
    tags=["endorse-transaction"],
    summary="For Endorser to refuse a particular transaction record",
)
@match_info_schema(TranIdMatchInfoSchema())
@response_schema(TransactionRecordSchema(), 200)
@tenant_authentication
async def refuse_transaction_response(request: web.BaseRequest):
    """Request handler for creating a refused transaction response.

    Args:
        request: aiohttp request object
    Returns:
        The updated transaction record details
    """

    context: AdminRequestContext = request["context"]
    outbound_handler = request["outbound_message_router"]

    transaction_id = request.match_info["tran_id"]
    try:
        async with context.profile.session() as session:
            transaction = await TransactionRecord.retrieve_by_id(session, transaction_id)
            connection_record = await ConnRecord.retrieve_by_id(
                session, transaction.connection_id
            )
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except BaseModelError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    async with context.profile.session() as session:
        jobs = await connection_record.metadata_get(session, "transaction_jobs")
    if not jobs:
        raise web.HTTPForbidden(
            reason=(
                "The transaction related jobs are not set up in "
                "connection metadata for this connection record"
            )
        )
    if jobs["transaction_my_job"] != TransactionJob.TRANSACTION_ENDORSER.name:
        raise web.HTTPForbidden(
            reason="Only a TRANSACTION_ENDORSER can refuse a transaction"
        )

    try:
        transaction_mgr = TransactionManager(context.profile)
        (
            transaction,
            refused_transaction_response,
        ) = await transaction_mgr.create_refuse_response(
            transaction=transaction,
            state=TransactionRecord.STATE_TRANSACTION_REFUSED,
            refuser_did=None,
        )
    except (StorageError, TransactionManagerError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    await outbound_handler(
        refused_transaction_response, connection_id=transaction.connection_id
    )

    return web.json_response(transaction.serialize())


@docs(
    tags=["endorse-transaction"],
    summary="For Author to cancel a particular transaction request",
)
@match_info_schema(TranIdMatchInfoSchema())
@response_schema(TransactionRecordSchema(), 200)
@tenant_authentication
async def cancel_transaction(request: web.BaseRequest):
    """Request handler for cancelling a Transaction request.

    Args:
        request: aiohttp request object
    Returns:
        The updated transaction record details
    """

    context: AdminRequestContext = request["context"]
    outbound_handler = request["outbound_message_router"]
    transaction_id = request.match_info["tran_id"]
    try:
        async with context.profile.session() as session:
            transaction = await TransactionRecord.retrieve_by_id(session, transaction_id)
            connection_record = await ConnRecord.retrieve_by_id(
                session, transaction.connection_id
            )
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except BaseModelError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    async with context.profile.session() as session:
        jobs = await connection_record.metadata_get(session, "transaction_jobs")
    if not jobs:
        raise web.HTTPForbidden(
            reason=(
                "The transaction related jobs are not set up in "
                "connection metadata for this connection record"
            )
        )
    if jobs["transaction_my_job"] != TransactionJob.TRANSACTION_AUTHOR.name:
        raise web.HTTPForbidden(
            reason="Only a TRANSACTION_AUTHOR can cancel a transaction"
        )

    transaction_mgr = TransactionManager(context.profile)
    try:
        (
            transaction,
            cancelled_transaction_response,
        ) = await transaction_mgr.cancel_transaction(
            transaction=transaction,
            state=TransactionRecord.STATE_TRANSACTION_CANCELLED,
        )
    except (StorageError, TransactionManagerError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    await outbound_handler(
        cancelled_transaction_response, connection_id=transaction.connection_id
    )

    return web.json_response(transaction.serialize())


@docs(
    tags=["endorse-transaction"],
    summary="For Author to resend a particular transaction request",
)
@match_info_schema(TranIdMatchInfoSchema())
@response_schema(TransactionRecordSchema(), 200)
@tenant_authentication
async def transaction_resend(request: web.BaseRequest):
    """Request handler for resending a transaction request.

    Args:
        request: aiohttp request object
    Returns:
        The updated transaction record details
    """

    context: AdminRequestContext = request["context"]
    outbound_handler = request["outbound_message_router"]
    transaction_id = request.match_info["tran_id"]
    try:
        async with context.profile.session() as session:
            transaction = await TransactionRecord.retrieve_by_id(session, transaction_id)
            connection_record = await ConnRecord.retrieve_by_id(
                session, transaction.connection_id
            )
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except BaseModelError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    async with context.profile.session() as session:
        jobs = await connection_record.metadata_get(session, "transaction_jobs")
    if not jobs:
        raise web.HTTPForbidden(
            reason=(
                "The transaction related jobs are not set up in "
                "connection metadata for this connection record"
            )
        )
    if jobs["transaction_my_job"] != TransactionJob.TRANSACTION_AUTHOR.name:
        raise web.HTTPForbidden(
            reason="Only a TRANSACTION_AUTHOR can resend a transaction"
        )

    try:
        transaction_mgr = TransactionManager(context.profile)
        (
            transaction,
            resend_transaction_response,
        ) = await transaction_mgr.transaction_resend(
            transaction=transaction, state=TransactionRecord.STATE_TRANSACTION_RESENT
        )
    except (StorageError, TransactionManagerError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    await outbound_handler(
        resend_transaction_response, connection_id=transaction.connection_id
    )

    return web.json_response(transaction.serialize())


@docs(
    tags=["endorse-transaction"],
    summary="Set transaction jobs",
)
@querystring_schema(AssignTransactionJobsSchema())
@match_info_schema(TransactionConnIdMatchInfoSchema())
@response_schema(TransactionJobsSchema(), 200)
@tenant_authentication
async def set_endorser_role(request: web.BaseRequest):
    """Request handler for assigning transaction jobs.

    Args:
        request: aiohttp request object
    Returns:
        The assigned transaction jobs
    """

    context: AdminRequestContext = request["context"]
    outbound_handler = request["outbound_message_router"]
    connection_id = request.match_info["conn_id"]
    transaction_my_job = request.query.get("transaction_my_job")

    async with context.profile.session() as session:
        try:
            record = await ConnRecord.retrieve_by_id(session, connection_id)
        except StorageNotFoundError as err:
            raise web.HTTPNotFound(reason=err.roll_up) from err
        except BaseModelError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

    transaction_mgr = TransactionManager(context.profile)
    tx_job_to_send = await transaction_mgr.set_transaction_my_job(
        record=record, transaction_my_job=transaction_my_job
    )
    async with context.profile.session() as session:
        jobs = await record.metadata_get(session, "transaction_jobs")

    await outbound_handler(tx_job_to_send, connection_id=connection_id)
    return web.json_response(jobs)


@docs(
    tags=["endorse-transaction"],
    summary="Set Endorser Info",
)
@querystring_schema(EndorserInfoSchema())
@match_info_schema(TransactionConnIdMatchInfoSchema())
@response_schema(EndorserInfoSchema(), 200)
@tenant_authentication
async def set_endorser_info(request: web.BaseRequest):
    """Request handler for assigning endorser information.

    Args:
        request: aiohttp request object
    Returns:
        The assigned endorser information
    """

    context: AdminRequestContext = request["context"]
    connection_id = request.match_info["conn_id"]
    endorser_did = request.query.get("endorser_did")
    endorser_name = request.query.get("endorser_name")

    async with context.profile.session() as session:
        try:
            record = await ConnRecord.retrieve_by_id(session, connection_id)
        except StorageNotFoundError as err:
            raise web.HTTPNotFound(reason=err.roll_up) from err
        except BaseModelError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err
        jobs = await record.metadata_get(session, "transaction_jobs")
    if not jobs:
        raise web.HTTPForbidden(
            reason=(
                "The transaction related jobs are not set up in "
                "connection metadata for this connection record"
            )
        )
    if "transaction_my_job" not in jobs.keys():
        raise web.HTTPForbidden(
            reason=(
                'The "transaction_my_job" is not set in "transaction_jobs"'
                " in connection metadata for this connection record"
            )
        )
    if jobs["transaction_my_job"] != TransactionJob.TRANSACTION_AUTHOR.name:
        raise web.HTTPForbidden(
            reason=(
                "Only a TRANSACTION_AUTHOR can add endorser_info "
                "to metadata of its connection record"
            )
        )
    async with context.profile.session() as session:
        value = await record.metadata_get(session, "endorser_info")
        if value:
            value["endorser_did"] = endorser_did
            value["endorser_name"] = endorser_name
        else:
            value = {"endorser_did": endorser_did, "endorser_name": endorser_name}
        await record.metadata_set(session, key="endorser_info", value=value)

        endorser_info = await record.metadata_get(session, "endorser_info")

    return web.json_response(endorser_info)


@docs(
    tags=["endorse-transaction"],
    summary="For Author / Endorser to write an endorsed transaction to the ledger",
)
@match_info_schema(TranIdMatchInfoSchema())
@response_schema(TransactionRecordSchema(), 200)
@tenant_authentication
async def transaction_write(request: web.BaseRequest):
    """Request handler for writing an endorsed transaction to the ledger.

    Args:
        request: aiohttp request object
    Returns:
        The returned ledger response
    """

    context: AdminRequestContext = request["context"]
    outbound_handler = request["outbound_message_router"]
    transaction_id = request.match_info["tran_id"]
    try:
        async with context.profile.session() as session:
            transaction = await TransactionRecord.retrieve_by_id(session, transaction_id)
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except BaseModelError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    if transaction.state != TransactionRecord.STATE_TRANSACTION_ENDORSED:
        raise web.HTTPForbidden(
            reason=(
                " The transaction cannot be written to the ledger as it is in state: "
            )
            + transaction.state
        )

    # update the final transaction status
    transaction_mgr = TransactionManager(context.profile)
    try:
        (
            tx_completed,
            transaction_acknowledgement_message,
        ) = await transaction_mgr.complete_transaction(transaction, False)
    except StorageError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    await outbound_handler(
        transaction_acknowledgement_message, connection_id=transaction.connection_id
    )

    return web.json_response(tx_completed.serialize())


def register_events(event_bus: EventBus):
    """Subscribe to any events we need to support."""
    event_bus.subscribe(STARTUP_EVENT_PATTERN, on_startup_event)
    event_bus.subscribe(SHUTDOWN_EVENT_PATTERN, on_shutdown_event)


async def on_startup_event(profile: Profile, event: Event):
    """Handle any events we need to support."""

    await attempt_auto_author_with_endorser_setup(profile)


async def on_shutdown_event(profile: Profile, event: Event):
    """Handle any events we need to support."""
    # nothing to do for now ...
    pass


async def register(app: web.Application):
    """Register routes."""

    app.add_routes(
        [
            web.get("/transactions", transactions_list, allow_head=False),
            web.get("/transactions/{tran_id}", transactions_retrieve, allow_head=False),
            web.post("/transactions/create-request", transaction_create_request),
            web.post("/transactions/{tran_id}/endorse", endorse_transaction_response),
            web.post("/transactions/{tran_id}/refuse", refuse_transaction_response),
            web.post("/transactions/{tran_id}/cancel", cancel_transaction),
            web.post("/transaction/{tran_id}/resend", transaction_resend),
            web.post("/transactions/{conn_id}/set-endorser-role", set_endorser_role),
            web.post("/transactions/{conn_id}/set-endorser-info", set_endorser_info),
            web.post("/transactions/{tran_id}/write", transaction_write),
        ]
    )


def post_process_routes(app: web.Application):
    """Amend swagger API."""

    # Add top-level tags description
    if "tags" not in app._state["swagger_dict"]:
        app._state["swagger_dict"]["tags"] = []
    app._state["swagger_dict"]["tags"].append(
        {
            "name": "endorse-transaction",
            "description": "Endorse a Transaction",
        }
    )



"""Basic message admin routes."""

from aiohttp import web
from aiohttp_apispec import docs, match_info_schema, request_schema, response_schema
from marshmallow import fields

from ....admin.decorators.auth import tenant_authentication
from ....admin.request_context import AdminRequestContext
from ....connections.models.conn_record import ConnRecord
from ....messaging.models.openapi import OpenAPISchema
from ....messaging.valid import UUID4_EXAMPLE
from ....storage.error import StorageNotFoundError
from .message_types import SPEC_URI
from .messages.basicmessage import BasicMessage


class BasicMessageModuleResponseSchema(OpenAPISchema):
    """Response schema for Basic Message Module."""


class SendMessageSchema(OpenAPISchema):
    """Request schema for sending a message."""

    content = fields.Str(metadata={"description": "Message content", "example": "Hello"})


class BasicConnIdMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking connection id."""

    conn_id = fields.Str(
        required=True,
        metadata={"description": "Connection identifier", "example": UUID4_EXAMPLE},
    )


@docs(tags=["basicmessage"], summary="Send a basic message to a connection")
@match_info_schema(BasicConnIdMatchInfoSchema())
@request_schema(SendMessageSchema())
@response_schema(BasicMessageModuleResponseSchema(), 200, description="")
@tenant_authentication
async def connections_send_message(request: web.BaseRequest):
    """Request handler for sending a basic message to a connection.

    Args:
        request: aiohttp request object

    """
    context: AdminRequestContext = request["context"]
    connection_id = request.match_info["conn_id"]
    outbound_handler = request["outbound_message_router"]
    params = await request.json()

    try:
        async with context.profile.session() as session:
            connection = await ConnRecord.retrieve_by_id(session, connection_id)
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err

    if connection.is_ready:
        msg = BasicMessage(content=params["content"])
        await outbound_handler(msg, connection_id=connection_id)

    return web.json_response({})


async def register(app: web.Application):
    """Register routes."""

    app.add_routes(
        [web.post("/connections/{conn_id}/send-message", connections_send_message)]
    )


def post_process_routes(app: web.Application):
    """Amend swagger API."""

    # Add top-level tags description
    if "tags" not in app._state["swagger_dict"]:
        app._state["swagger_dict"]["tags"] = []
    app._state["swagger_dict"]["tags"].append(
        {
            "name": "basicmessage",
            "description": "Simple messaging",
            "externalDocs": {"description": "Specification", "url": SPEC_URI},
        }
    )



"""A problem report message."""

from enum import Enum

from marshmallow import EXCLUDE, ValidationError, validates_schema

from ....problem_report.v1_0.message import ProblemReport, ProblemReportSchema
from ..message_types import PRES_20_PROBLEM_REPORT, PROTOCOL_PACKAGE

HANDLER_CLASS = (
    f"{PROTOCOL_PACKAGE}.handlers.pres_problem_report_handler.V20PresProblemReportHandler"
)


class ProblemReportReason(Enum):
    """Supported reason codes."""

    ABANDONED = "abandoned"


class V20PresProblemReport(ProblemReport):
    """Class representing a problem report message."""

    class Meta:
        """Problem report metadata."""

        handler_class = HANDLER_CLASS
        schema_class = "V20PresProblemReportSchema"
        message_type = PRES_20_PROBLEM_REPORT

    def __init__(self, *args, **kwargs):
        """Initialize problem report object."""
        super().__init__(*args, **kwargs)


class V20PresProblemReportSchema(ProblemReportSchema):
    """Problem report schema."""

    class Meta:
        """Schema metadata."""

        model_class = V20PresProblemReport
        unknown = EXCLUDE

    @validates_schema
    def validate_fields(self, data, **kwargs):
        """Validate schema fields.

        Args:
            data: The data to validate
            kwargs: Additional keyword arguments

        """
        if not data.get("description", {}).get("code", ""):
            raise ValidationError("Value for description.code must be present")



"""Classes for BaseStorage-based record management."""

import json
import logging
import sys
from datetime import datetime
from typing import Any, Mapping, Optional, Sequence, Type, TypeVar, Union

from marshmallow import fields
from uuid_utils import uuid4

from ...cache.base import BaseCache
from ...config.settings import BaseSettings
from ...core.profile import ProfileSession
from ...storage.base import (
    DEFAULT_PAGE_SIZE,
    BaseStorage,
    StorageDuplicateError,
    StorageNotFoundError,
)
from ...storage.record import StorageRecord
from ..util import datetime_to_str, time_now
from ..valid import ISO8601_DATETIME_EXAMPLE, ISO8601_DATETIME_VALIDATE
from .base import BaseModel, BaseModelError, BaseModelSchema

LOGGER = logging.getLogger(__name__)


RecordType = TypeVar("RecordType", bound="BaseRecord")


def match_post_filter(
    record: dict,
    post_filter: dict,
    positive: bool = True,
    alt: bool = False,
) -> bool:
    """Determine if a record value matches the post-filter.

    Args:
        record: record to check
        post_filter: filter to apply (empty or None filter matches everything)
        positive: whether matching all filter criteria positively or negatively
        alt: set to match any (positive=True) value or miss all (positive=False)
            values in post_filter
    """
    if not post_filter:
        return True

    if alt:
        return (
            positive
            and all(
                record.get(k) and record.get(k) in alts for k, alts in post_filter.items()
            )
        ) or (
            (not positive)
            and all(
                record.get(k) and record.get(k) not in alts
                for k, alts in post_filter.items()
            )
        )

    for k, v in post_filter.items():
        if record.get(k) != v:
            return not positive

    return positive


class BaseRecord(BaseModel):
    """Represents a single storage record."""

    class Meta:
        """BaseRecord metadata."""

    DEFAULT_CACHE_TTL = 60
    RECORD_ID_NAME = "id"
    RECORD_TYPE = None
    RECORD_TOPIC: Optional[str] = None
    EVENT_NAMESPACE: str = "acapy::record"
    LOG_STATE_FLAG = None
    TAG_NAMES = {"state"}
    STATE_DELETED = "deleted"

    def __init__(
        self,
        id: Optional[str] = None,
        state: Optional[str] = None,
        *,
        created_at: Union[str, datetime, None] = None,
        updated_at: Union[str, datetime, None] = None,
        new_with_id: bool = False,
    ):
        """Initialize a new BaseRecord."""
        if not self.RECORD_TYPE:
            raise TypeError(
                "Cannot instantiate abstract class {} with no RECORD_TYPE".format(
                    self.__class__.__name__
                )
            )
        self._id = id
        self._last_state = state
        self._new_with_id = new_with_id
        self.state = state
        self.created_at = datetime_to_str(created_at)
        self.updated_at = datetime_to_str(updated_at)

    @classmethod
    def from_storage(cls, record_id: str, record: Mapping[str, Any]):
        """Initialize a record from its stored representation.

        Args:
            record_id: The unique record identifier
            record: The stored representation
        """
        record_id_name = cls.RECORD_ID_NAME
        if record_id_name in record:
            raise ValueError(f"Duplicate {record_id_name} inputs; {record}")
        params = dict(**record)
        params[record_id_name] = record_id
        return cls(**params)

    @classmethod
    def get_tag_map(cls) -> Mapping[str, str]:
        """Accessor for the set of defined tags."""

        return {tag.lstrip("~"): tag for tag in cls.TAG_NAMES or ()}

    @property
    def storage_record(self) -> StorageRecord:
        """Accessor for a `StorageRecord` representing this record."""

        return StorageRecord(
            self.RECORD_TYPE, json.dumps(self.value), self.tags, self._id
        )

    @property
    def record_value(self) -> dict:
        """Accessor to define custom properties for the JSON record value."""

        return {}

    @property
    def value(self) -> dict:
        """Accessor for the JSON record value generated for this record."""

        ret = self.strip_tag_prefix(self.tags)
        ret.update({"created_at": self.created_at, "updated_at": self.updated_at})
        ret.update(self.record_value)
        return ret

    @property
    def record_tags(self) -> dict:
        """Accessor to define implementation-specific tags."""

        return {
            tag: getattr(self, prop)
            for (prop, tag) in self.get_tag_map().items()
            if getattr(self, prop) is not None
        }

    @property
    def tags(self) -> dict:
        """Accessor for the record tags generated for this record."""

        tags = self.record_tags
        return tags

    @classmethod
    async def get_cached_key(cls, session: ProfileSession, cache_key: str):
        """Shortcut method to fetch a cached key value.

        Args:
            session: The profile session to use
            cache_key: The unique cache identifier
        """
        if not cache_key:
            return
        cache = session.inject_or(BaseCache)
        if cache:
            return await cache.get(cache_key)

    @classmethod
    async def set_cached_key(
        cls, session: ProfileSession, cache_key: str, value: Any, ttl=None
    ):
        """Shortcut method to set a cached key value.

        Args:
            session: The profile session to use
            cache_key: The unique cache identifier
            value: The value to cache
            ttl: The cache ttl
        """

        if not cache_key:
            return
        cache = session.inject_or(BaseCache)
        if cache:
            await cache.set(cache_key, value, ttl or cls.DEFAULT_CACHE_TTL)

    @classmethod
    async def clear_cached_key(cls, session: ProfileSession, cache_key: str):
        """Shortcut method to clear a cached key value, if any.

        Args:
            session: The profile session to use
            cache_key: The unique cache identifier
        """

        if not cache_key:
            return
        cache = session.inject_or(BaseCache)
        if cache:
            await cache.clear(cache_key)

    @classmethod
    async def retrieve_by_id(
        cls: Type[RecordType],
        session: ProfileSession,
        record_id: str,
        *,
        for_update=False,
    ) -> RecordType:
        """Retrieve a stored record by ID.

        Args:
            session: The profile session to use
            record_id: The ID of the record to find
            for_update: Whether to lock the record for update
        """

        storage = session.inject(BaseStorage)
        result = await storage.get_record(
            cls.RECORD_TYPE, record_id, options={"forUpdate": for_update}
        )
        vals = json.loads(result.value)
        return cls.from_storage(record_id, vals)

    @classmethod
    async def retrieve_by_tag_filter(
        cls: Type[RecordType],
        session: ProfileSession,
        tag_filter: dict,
        post_filter: Optional[dict] = None,
        *,
        for_update=False,
    ) -> RecordType:
        """Retrieve a record by tag filter.

        Args:
            cls: The record class
            session: The profile session to use
            tag_filter: The filter dictionary to apply
            post_filter: Additional value filters to apply matching positively,
                with sequence values specifying alternatives to match (hit any)
            for_update: Whether to lock the record for update
        """

        storage = session.inject(BaseStorage)
        rows = await storage.find_all_records(
            cls.RECORD_TYPE,
            cls.prefix_tag_filter(tag_filter),
            options={"forUpdate": for_update},
        )
        found = None
        for record in rows:
            vals = json.loads(record.value)
            if match_post_filter(vals, post_filter, alt=False):
                if found:
                    raise StorageDuplicateError(
                        "Multiple {} records located for {}{}".format(
                            cls.__name__,
                            tag_filter,
                            f", {post_filter}" if post_filter else "",
                        )
                    )
                found = cls.from_storage(record.id, vals)
        if not found:
            raise StorageNotFoundError(
                "{} record not found for {}{}".format(
                    cls.__name__, tag_filter, f", {post_filter}" if post_filter else ""
                )
            )
        return found

    @classmethod
    async def query(
        cls: Type[RecordType],
        session: ProfileSession,
        tag_filter: Optional[dict] = None,
        *,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        order_by: Optional[str] = None,
        descending: bool = False,
        post_filter_positive: Optional[dict] = None,
        post_filter_negative: Optional[dict] = None,
        alt: bool = False,
    ) -> Sequence[RecordType]:
        """Query stored records.

        Args:
            session: The profile session to use
            tag_filter: An optional dictionary of tag filter clauses
            limit: The maximum number of records to retrieve
            offset: The offset to start retrieving records from
            order_by: An optional field by which to order the records.
            descending: Whether to order the records in descending order.
            post_filter_positive: Additional value filters to apply matching positively
            post_filter_negative: Additional value filters to apply matching negatively
            alt: set to match any (positive=True) value or miss all (positive=False)
                values in post_filter
        """

        storage = session.inject(BaseStorage)

        tag_query = cls.prefix_tag_filter(tag_filter)
        post_filter = post_filter_positive or post_filter_negative

        # set flag to indicate if pagination is requested or not, then set defaults
        paginated = limit is not None or offset is not None
        limit = limit or DEFAULT_PAGE_SIZE
        offset = offset or 0

        if not post_filter and paginated:
            # Only fetch paginated records if post-filter is not being applied
            rows = await storage.find_paginated_records(
                type_filter=cls.RECORD_TYPE,
                tag_query=tag_query,
                limit=limit,
                offset=offset,
                order_by=order_by,
                descending=descending,
            )
        else:
            rows = await storage.find_all_records(
                type_filter=cls.RECORD_TYPE,
                tag_query=tag_query,
                order_by=order_by,
                descending=descending,
            )

        num_results_post_filter = 0  # used if applying pagination post-filter
        num_records_to_match = limit + offset  # ignored if not paginated

        result = []
        for record in rows:
            try:
                vals = json.loads(record.value)
                if not post_filter:  # pagination would already be applied if requested
                    result.append(cls.from_storage(record.id, vals))
                else:
                    continue_processing = (
                        not paginated or num_results_post_filter < num_records_to_match
                    )
                    if not continue_processing:
                        break

                    post_filter_match = match_post_filter(
                        vals, post_filter_positive, positive=True, alt=alt
                    ) and match_post_filter(
                        vals, post_filter_negative, positive=False, alt=alt
                    )

                    if not post_filter_match:
                        continue

                    if num_results_post_filter >= offset:  # append only after offset
                        result.append(cls.from_storage(record.id, vals))

                    num_results_post_filter += 1
            except (BaseModelError, json.JSONDecodeError, TypeError) as err:
                raise BaseModelError(f"{err}, for record id {record.id}")
        return result

    async def save(
        self,
        session: ProfileSession,
        *,
        reason: Optional[str] = None,
        log_params: Mapping[str, Any] = None,
        log_override: bool = False,
        event: Optional[bool] = None,
    ) -> str:
        """Persist the record to storage.

        Args:
            session: The profile session to use
            reason: A reason to add to the log
            log_params: Additional parameters to log
            log_override: Override configured logging regimen, print to stderr instead
            event: Flag to override whether the event is sent
        """

        new_record = None
        log_reason = reason or ("Updated record" if self._id else "Created record")
        try:
            self.updated_at = time_now()
            storage = session.inject(BaseStorage)
            if self._id and not self._new_with_id:
                record = self.storage_record
                await storage.update_record(record, record.value, record.tags)
                new_record = False
            else:
                if not self._id:
                    self._id = str(uuid4())
                self.created_at = self.updated_at
                await storage.add_record(self.storage_record)
                new_record = True
                self._new_with_id = False
        finally:
            params = {self.RECORD_TYPE: self.serialize()}
            if log_params:
                params.update(log_params)
            if new_record is None:
                log_reason = f"FAILED: {log_reason}"
            self.log_state(
                log_reason, params, override=log_override, settings=session.settings
            )

        await self.post_save(session, new_record, self._last_state, event)
        self._last_state = self.state

        return self._id

    async def post_save(
        self,
        session: ProfileSession,
        new_record: bool,
        last_state: Optional[str],
        event: Optional[bool] = None,
    ):
        """Perform post-save actions.

        Args:
            session: The profile session to use
            new_record: Flag indicating if the record was just created
            last_state: The previous state value
            event: Flag to override whether the event is sent
        """

        if event is None:
            event = new_record or (last_state != self.state)
        if event:
            await self.emit_event(session, self.serialize())

    async def delete_record(self, session: ProfileSession):
        """Remove the stored record.

        Args:
            session: The profile session to use
        """

        if self._id:
            storage = session.inject(BaseStorage)
            if self.state:
                self._previous_state = self.state
                self.state = BaseRecord.STATE_DELETED
                await self.emit_event(session, self.serialize())
            await storage.delete_record(self.storage_record)

    async def emit_event(self, session: ProfileSession, payload: Optional[Any] = None):
        """Emit an event.

        Args:
            session: The profile session to use
            payload: The event payload
        """

        if not self.RECORD_TOPIC:
            return

        if self.state:
            topic = f"{self.EVENT_NAMESPACE}::{self.RECORD_TOPIC}::{self.state}"
        else:
            topic = f"{self.EVENT_NAMESPACE}::{self.RECORD_TOPIC}"

        if not payload:
            payload = self.serialize()

        await session.emit_event(topic, payload)

    @classmethod
    def log_state(
        cls,
        msg: str,
        params: Optional[dict] = None,
        settings: Optional[BaseSettings] = None,
        override: bool = False,
    ):
        """Print a message with increased visibility (for testing)."""

        if override or (
            cls.LOG_STATE_FLAG and settings and settings.get(cls.LOG_STATE_FLAG)
        ):
            out = msg + "\n"
            if params:
                for k, v in params.items():
                    out += f"    {k}: {v}\n"
            print(out, file=sys.stderr)

    @classmethod
    def strip_tag_prefix(cls, tags: dict):
        """Strip tilde from unencrypted tag names."""

        return {(k[1:] if "~" in k else k): v for (k, v) in tags.items()} if tags else {}

    @classmethod
    def prefix_tag_filter(cls, tag_filter: dict):
        """Prefix unencrypted tags used in the tag filter."""

        ret = None
        if tag_filter:
            tag_map = cls.get_tag_map()
            ret = {}
            for k, v in tag_filter.items():
                if k in ("$or", "$and") and isinstance(v, list):
                    ret[k] = [cls.prefix_tag_filter(clause) for clause in v]
                elif k == "$not" and isinstance(v, dict):
                    ret[k] = cls.prefix_tag_filter(v)
                else:
                    ret[tag_map.get(k, k)] = v
        return ret

    def __eq__(self, other: Any) -> bool:
        """Comparison between records."""

        if type(other) is type(self):
            return self.value == other.value and self.tags == other.tags
        return False

    @classmethod
    def get_attributes_by_prefix(cls, prefix: str, walk_mro: bool = True):
        """List all values for attributes with common prefix.

        Args:
            prefix: Common prefix to look for
            walk_mro: Walk MRO to find attributes inherited from superclasses
        """

        bases = cls.__mro__ if walk_mro else [cls]
        return [
            vars(base)[name]
            for base in bases
            for name in vars(base)
            if name.startswith(prefix)
        ]


class BaseExchangeRecord(BaseRecord):
    """Represents a base record with event tracing capability."""

    def __init__(
        self,
        id: Optional[str] = None,
        state: Optional[str] = None,
        *,
        trace: bool = False,
        **kwargs,
    ):
        """Initialize a new BaseExchangeRecord."""

        super().__init__(id, state, **kwargs)
        self.trace = trace

    def __eq__(self, other: Any) -> bool:
        """Comparison between records."""

        if type(other) is type(self):
            return (
                self.value == other.value
                and self.tags == other.tags
                and self.trace == other.trace
            )
        return False


class BaseRecordSchema(BaseModelSchema):
    """Schema to allow serialization/deserialization of base records."""

    class Meta:
        """BaseRecordSchema metadata."""

        model_class = None

    state = fields.Str(
        required=False,
        metadata={"description": "Current record state", "example": "active"},
    )
    created_at = fields.Str(
        required=False,
        validate=ISO8601_DATETIME_VALIDATE,
        metadata={
            "description": "Time of record creation",
            "example": ISO8601_DATETIME_EXAMPLE,
        },
    )
    updated_at = fields.Str(
        required=False,
        validate=ISO8601_DATETIME_VALIDATE,
        metadata={
            "description": "Time of last record update",
            "example": ISO8601_DATETIME_EXAMPLE,
        },
    )


class BaseExchangeSchema(BaseRecordSchema):
    """Base schema for exchange records."""

    class Meta:
        """BaseExchangeSchema metadata."""

        model_class = BaseExchangeRecord

    trace = fields.Boolean(
        required=False,
        dump_default=False,
        metadata={
            "description": "Record trace information, based on agent configuration"
        },
    )



"""Represents an response to a trust ping message."""

from typing import Optional

from marshmallow import EXCLUDE, fields

from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from ..message_types import PING_RESPONSE, PROTOCOL_PACKAGE

HANDLER_CLASS = f"{PROTOCOL_PACKAGE}.handlers.ping_response_handler.PingResponseHandler"


class PingResponse(AgentMessage):
    """Class representing a ping response."""

    class Meta:
        """PingResponse metadata."""

        handler_class = HANDLER_CLASS
        message_type = PING_RESPONSE
        schema_class = "PingResponseSchema"

    def __init__(self, *, comment: Optional[str] = None, **kwargs):
        """Initialize a PingResponse message instance.

        Args:
            comment: An optional comment string to include in the message
            kwargs: Keyword arguments for message construction

        """
        super().__init__(**kwargs)
        self.comment = comment


class PingResponseSchema(AgentMessageSchema):
    """PingResponse schema."""

    class Meta:
        """PingResponseSchema metadata."""

        model_class = PingResponse
        unknown = EXCLUDE

    comment = fields.Str(
        required=False,
        allow_none=True,
        metadata={"description": "Optional comment to include", "example": "Hello"},
    )



from unittest import TestCase

import pytest

from ..bbs import (
    BbsException,
    create_bls12381g2_keypair,
    sign_messages_bls12381g2,
    verify_signed_messages_bls12381g2,
)

PUBLIC_KEY_BYTES = b"\x8c\xa2\xbe \xad\x9c\xf9\x98\x04e\x17\xec\x00\xf9J\xd7\xed\xdb\xa4\x0el\x9fG\x94\x9b+d\xac\xf5\x0f{\xf3\xa1\xc7\xd9\xa7*YZ\xa78\x86\x8b\xcc\xb8\x9d\x86\xcb\x07\x8b\xfak9\xd8\xb8\\\x93\x86\xc8\xf7\xb1O\x88\x17\xa4\xc9x5\xdd\x04e\xee'\xff\x05\xf8&EU\xfa\x8f\x83\xa3#\xca\x9b\xdf\xa8\xc2\xda\x0fz\x8f\x05@o"
SECRET_KEY_BYTES = b"TmL\x10\xd4k[\x8b\xc3\xb7\x91\xd5\x90\xaf4\xbe\xdc\x89f\xb4\r\xbeV\xbdq>\xbd=\xee\x97\x86\x01"
SIGNED_BYTES = b"\x82\xd7\x9c\xa9\x05\xe1\xb8\x0c\x92\xa1Jr\xd0G%\x18`P\x0e\xa8\xf6\xcapWe\xa0\xc0\xa1c\x92Tr\xb6k\xf1\xa5\x196\xbd\xa4\x94\xe0\xeatO$Re>\x10\x05[?\x93\x8f\x03\x8d\xa1\x11r\x98\xd5\x17\xf4\xe4\xb5\x80\xcbh\xfc_\x8cy\xe8p\x98\x1d\xcd\xeb+*\xf7E\xa1\x10\xf3\xa5\x8f^\xda\xba\xc5&\xb9\xdfI\xe6bl\xdcx\x7fM\xc6\xbf,eA|\x99h\xcd"
SIGN_MESSAGES = [b"messag1", b"message2"]
SEED = "seed000000000001"


@pytest.mark.ursa_bbs_signatures
class TestBBS(TestCase):
    def test_create_keypair_seed(self):
        (pk, sk) = create_bls12381g2_keypair(SEED)

        assert pk == PUBLIC_KEY_BYTES
        assert sk == SECRET_KEY_BYTES

    def test_create_keypair(self):
        (pk, sk) = create_bls12381g2_keypair()

        assert pk
        assert sk

    def test_create_keypair_x_invalid_seed(self):
        with self.assertRaises(BbsException) as context:
            create_bls12381g2_keypair(10)
        assert "Unable to create keypair" in str(context.exception)

    def test_sign(self):
        signed = sign_messages_bls12381g2(SIGN_MESSAGES, SECRET_KEY_BYTES)

        assert signed

        assert verify_signed_messages_bls12381g2(SIGN_MESSAGES, signed, PUBLIC_KEY_BYTES)

    def test_sign_x_invalid_secret_key_bytes(self):
        with self.assertRaises(BbsException) as context:
            sign_messages_bls12381g2(SIGN_MESSAGES, "hello")
        assert "Unable to sign messages" in str(context.exception)

    def test_verify(self):
        assert verify_signed_messages_bls12381g2(
            SIGN_MESSAGES, SIGNED_BYTES, PUBLIC_KEY_BYTES
        )

    def test_verify_x_invalid_pk(self):
        with self.assertRaises(BbsException):
            verify_signed_messages_bls12381g2(
                SIGN_MESSAGES, SIGNED_BYTES, PUBLIC_KEY_BYTES + b"10"
            )

    def test_verify_x_invalid_messages(self):
        with self.assertRaises(BbsException):
            verify_signed_messages_bls12381g2(
                SIGN_MESSAGES, SIGNED_BYTES, PUBLIC_KEY_BYTES + b"10"
            )
        assert not verify_signed_messages_bls12381g2(
            [SIGN_MESSAGES[0]], SIGNED_BYTES, PUBLIC_KEY_BYTES
        )

    def test_verify_x_invalid_signed_bytes(self):
        with self.assertRaises(BbsException):
            assert not verify_signed_messages_bls12381g2(
                SIGN_MESSAGES, SIGNED_BYTES + b"10", PUBLIC_KEY_BYTES
            )



"""Credential schema admin routes."""

import json
from asyncio import shield
from time import time

from aiohttp import web
from aiohttp_apispec import (
    docs,
    match_info_schema,
    querystring_schema,
    request_schema,
    response_schema,
)
from marshmallow import fields
from marshmallow.validate import Regexp

from ...admin.decorators.auth import tenant_authentication
from ...admin.request_context import AdminRequestContext
from ...connections.models.conn_record import ConnRecord
from ...core.event_bus import Event, EventBus
from ...core.profile import Profile
from ...indy.issuer import IndyIssuer, IndyIssuerError
from ...indy.models.schema import SchemaSchema
from ...ledger.base import BaseLedger
from ...ledger.error import LedgerError
from ...ledger.multiple_ledger.ledger_requests_executor import (
    GET_SCHEMA,
    IndyLedgerRequestsExecutor,
)
from ...multitenant.base import BaseMultitenantManager
from ...protocols.endorse_transaction.v1_0.manager import (
    TransactionManager,
    TransactionManagerError,
)
from ...protocols.endorse_transaction.v1_0.models.transaction_record import (
    TransactionRecordSchema,
)
from ...protocols.endorse_transaction.v1_0.util import (
    get_endorser_connection_id,
    is_author_role,
)
from ...storage.base import BaseStorage, StorageRecord
from ...storage.error import StorageError, StorageNotFoundError
from ...utils.profiles import is_anoncreds_profile_raise_web_exception
from ..models.base import BaseModelError
from ..models.openapi import OpenAPISchema
from ..valid import (
    B58,
    INDY_SCHEMA_ID_EXAMPLE,
    INDY_SCHEMA_ID_VALIDATE,
    MAJOR_MINOR_VERSION_EXAMPLE,
    MAJOR_MINOR_VERSION_VALIDATE,
    UUID4_EXAMPLE,
)
from .util import (
    EVENT_LISTENER_PATTERN,
    SCHEMA_SENT_RECORD_TYPE,
    SCHEMA_TAGS,
    SchemaQueryStringSchema,
    notify_schema_event,
)


class SchemaSendRequestSchema(OpenAPISchema):
    """Request schema for schema send request."""

    schema_name = fields.Str(
        required=True, metadata={"description": "Schema name", "example": "prefs"}
    )
    schema_version = fields.Str(
        required=True,
        validate=MAJOR_MINOR_VERSION_VALIDATE,
        metadata={
            "description": "Schema version",
            "example": MAJOR_MINOR_VERSION_EXAMPLE,
        },
    )
    attributes = fields.List(
        fields.Str(metadata={"description": "attribute name", "example": "score"}),
        required=True,
        metadata={"description": "List of schema attributes"},
    )


class SchemaSendResultSchema(OpenAPISchema):
    """Result schema content for schema send request with auto-endorse."""

    schema_id = fields.Str(
        required=True,
        validate=INDY_SCHEMA_ID_VALIDATE,
        metadata={
            "description": "Schema identifier",
            "example": INDY_SCHEMA_ID_EXAMPLE,
        },
    )
    schema = fields.Nested(SchemaSchema(), metadata={"description": "Schema definition"})


class TxnOrSchemaSendResultSchema(OpenAPISchema):
    """Result schema for schema send request."""

    sent = fields.Nested(
        SchemaSendResultSchema(),
        required=False,
        metadata={"description": "Content sent"},
    )
    txn = fields.Nested(
        TransactionRecordSchema(),
        required=False,
        metadata={"description": "Schema transaction to endorse"},
    )


class SchemaGetResultSchema(OpenAPISchema):
    """Result schema for schema get request."""

    schema = fields.Nested(SchemaSchema())


class SchemasCreatedResultSchema(OpenAPISchema):
    """Result schema for a schemas-created request."""

    schema_ids = fields.List(
        fields.Str(
            validate=INDY_SCHEMA_ID_VALIDATE,
            metadata={
                "description": "Schema identifiers",
                "example": INDY_SCHEMA_ID_EXAMPLE,
            },
        )
    )


class SchemaIdMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking schema id."""

    schema_id = fields.Str(
        required=True,
        validate=Regexp(f"^[1-9][0-9]*|[{B58}]{{21,22}}:2:.+:[0-9.]+$"),
        metadata={
            "description": "Schema identifier",
            "example": INDY_SCHEMA_ID_EXAMPLE,
        },
    )


class CreateSchemaTxnForEndorserOptionSchema(OpenAPISchema):
    """Class for user to input whether to create a transaction for endorser or not."""

    create_transaction_for_endorser = fields.Boolean(
        required=False,
        metadata={"description": "Create Transaction For Endorser's signature"},
    )


class SchemaConnIdMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking connection id."""

    conn_id = fields.Str(
        required=False,
        metadata={"description": "Connection identifier", "example": UUID4_EXAMPLE},
    )


@docs(tags=["schema"], summary="Sends a schema to the ledger")
@request_schema(SchemaSendRequestSchema())
@querystring_schema(CreateSchemaTxnForEndorserOptionSchema())
@querystring_schema(SchemaConnIdMatchInfoSchema())
@response_schema(TxnOrSchemaSendResultSchema(), 200, description="")
@tenant_authentication
async def schemas_send_schema(request: web.BaseRequest):
    """Request handler for creating a schema.

    Args:
        request: aiohttp request object

    Returns:
        The schema id sent

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_anoncreds_profile_raise_web_exception(profile)

    outbound_handler = request["outbound_message_router"]

    create_transaction_for_endorser = json.loads(
        request.query.get("create_transaction_for_endorser", "false")
    )
    write_ledger = not create_transaction_for_endorser
    endorser_did = None
    connection_id = request.query.get("conn_id")

    body = await request.json()

    schema_name = body.get("schema_name")
    schema_version = body.get("schema_version")
    attributes = body.get("attributes")

    tag_query = {"schema_name": schema_name, "schema_version": schema_version}
    async with profile.session() as session:
        storage = session.inject(BaseStorage)
        found = await storage.find_all_records(
            type_filter=SCHEMA_SENT_RECORD_TYPE,
            tag_query=tag_query,
        )
        if 0 < len(found):
            raise web.HTTPBadRequest(
                reason=f"Schema {schema_name} {schema_version} already exists"
            )

    # check if we need to endorse
    if is_author_role(context.profile):
        # authors cannot write to the ledger
        write_ledger = False
        create_transaction_for_endorser = True
        if not connection_id:
            # author has not provided a connection id, so determine which to use
            connection_id = await get_endorser_connection_id(context.profile)
            if not connection_id:
                raise web.HTTPBadRequest(reason="No endorser connection found")

    if not write_ledger:
        try:
            async with profile.session() as session:
                connection_record = await ConnRecord.retrieve_by_id(
                    session, connection_id
                )
        except StorageNotFoundError as err:
            raise web.HTTPNotFound(reason=err.roll_up) from err
        except BaseModelError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

        async with profile.session() as session:
            endorser_info = await connection_record.metadata_get(session, "endorser_info")
        if not endorser_info:
            raise web.HTTPForbidden(
                reason=(
                    "Endorser Info is not set up in "
                    "connection metadata for this connection record"
                )
            )
        if "endorser_did" not in endorser_info.keys():
            raise web.HTTPForbidden(
                reason=(
                    ' "endorser_did" is not set in "endorser_info"'
                    " in connection metadata for this connection record"
                )
            )
        endorser_did = endorser_info["endorser_did"]

    ledger = context.inject_or(BaseLedger)
    if not ledger:
        reason = "No ledger available"
        if not context.settings.get_value("wallet.type"):
            reason += ": missing wallet-type?"
        raise web.HTTPForbidden(reason=reason)

    issuer = context.inject(IndyIssuer)
    async with ledger:
        try:
            # if create_transaction_for_endorser, then the returned "schema_def"
            # is actually the signed transaction
            schema_id, schema_def = await shield(
                ledger.create_and_send_schema(
                    issuer,
                    schema_name,
                    schema_version,
                    attributes,
                    write_ledger=write_ledger,
                    endorser_did=endorser_did,
                )
            )
        except (IndyIssuerError, LedgerError) as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

    meta_data = {
        "context": {
            "schema_id": schema_id,
            "schema_name": schema_name,
            "schema_version": schema_version,
            "attributes": attributes,
        },
        "processing": {},
    }

    if not create_transaction_for_endorser:
        # Notify event
        await notify_schema_event(context.profile, schema_id, meta_data)
        return web.json_response(
            {
                "sent": {"schema_id": schema_id, "schema": schema_def},
                "schema_id": schema_id,
                "schema": schema_def,
            }
        )

    # If the transaction is for the endorser, but the schema has already been created,
    # then we send back the schema since the transaction will fail to be created.
    elif "signed_txn" not in schema_def:
        return web.json_response({"sent": {"schema_id": schema_id, "schema": schema_def}})
    else:
        transaction_mgr = TransactionManager(context.profile)
        try:
            transaction = await transaction_mgr.create_record(
                messages_attach=schema_def["signed_txn"],
                connection_id=connection_id,
                meta_data=meta_data,
            )
        except StorageError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

        # if auto-request, send the request to the endorser
        if context.settings.get_value("endorser.auto_request"):
            try:
                transaction, transaction_request = await transaction_mgr.create_request(
                    transaction=transaction,
                    # TODO see if we need to parametrize these params
                    # expires_time=expires_time,
                )
            except (StorageError, TransactionManagerError) as err:
                raise web.HTTPBadRequest(reason=err.roll_up) from err

            await outbound_handler(transaction_request, connection_id=connection_id)

        return web.json_response(
            {
                "sent": {"schema_id": schema_id, "schema": schema_def},
                "txn": transaction.serialize(),
            }
        )


@docs(
    tags=["schema"],
    summary="Search for matching schema that agent originated",
)
@querystring_schema(SchemaQueryStringSchema())
@response_schema(SchemasCreatedResultSchema(), 200, description="")
@tenant_authentication
async def schemas_created(request: web.BaseRequest):
    """Request handler for retrieving schemas that current agent created.

    Args:
        request: aiohttp request object

    Returns:
        The identifiers of matching schemas

    """
    context: AdminRequestContext = request["context"]

    is_anoncreds_profile_raise_web_exception(context.profile)

    session = await context.session()
    storage = session.inject(BaseStorage)
    found = await storage.find_all_records(
        type_filter=SCHEMA_SENT_RECORD_TYPE,
        tag_query={
            tag: request.query[tag] for tag in SCHEMA_TAGS if tag in request.query
        },
    )

    return web.json_response({"schema_ids": [record.value for record in found]})


@docs(tags=["schema"], summary="Gets a schema from the ledger")
@match_info_schema(SchemaIdMatchInfoSchema())
@response_schema(SchemaGetResultSchema(), 200, description="")
@tenant_authentication
async def schemas_get_schema(request: web.BaseRequest):
    """Request handler for sending a credential offer.

    Args:
        request: aiohttp request object

    Returns:
        The schema details.

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_anoncreds_profile_raise_web_exception(profile)

    schema_id = request.match_info["schema_id"]

    async with profile.session() as session:
        multitenant_mgr = session.inject_or(BaseMultitenantManager)
        if multitenant_mgr:
            ledger_exec_inst = IndyLedgerRequestsExecutor(profile)
        else:
            ledger_exec_inst = session.inject(IndyLedgerRequestsExecutor)
    ledger_id, ledger = await ledger_exec_inst.get_ledger_for_identifier(
        schema_id,
        txn_record_type=GET_SCHEMA,
    )
    if not ledger:
        reason = "No ledger available"
        if not context.settings.get_value("wallet.type"):
            reason += ": missing wallet-type?"
        raise web.HTTPForbidden(reason=reason)

    async with ledger:
        try:
            schema = await ledger.get_schema(schema_id)
            if not schema:
                raise web.HTTPNotFound(reason=f"Schema not found: {schema_id}")
        except LedgerError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

    if ledger_id:
        return web.json_response({"ledger_id": ledger_id, "schema": schema})
    else:
        return web.json_response({"schema": schema})


@docs(tags=["schema"], summary="Writes a schema non-secret record to the wallet")
@match_info_schema(SchemaIdMatchInfoSchema())
@response_schema(SchemaGetResultSchema(), 200, description="")
@tenant_authentication
async def schemas_fix_schema_wallet_record(request: web.BaseRequest):
    """Request handler for fixing a schema's wallet non-secrets records.

    Args:
        request: aiohttp request object

    Returns:
        The schema details.

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    is_anoncreds_profile_raise_web_exception(profile)

    schema_id = request.match_info["schema_id"]

    async with profile.session() as session:
        storage = session.inject(BaseStorage)
        multitenant_mgr = session.inject_or(BaseMultitenantManager)
        if multitenant_mgr:
            ledger_exec_inst = IndyLedgerRequestsExecutor(profile)
        else:
            ledger_exec_inst = session.inject(IndyLedgerRequestsExecutor)
    ledger_id, ledger = await ledger_exec_inst.get_ledger_for_identifier(
        schema_id,
        txn_record_type=GET_SCHEMA,
    )
    if not ledger:
        reason = "No ledger available"
        if not context.settings.get_value("wallet.type"):
            reason += ": missing wallet-type?"
        raise web.HTTPForbidden(reason=reason)

    async with ledger:
        try:
            schema = await ledger.get_schema(schema_id)

            # check if the record exists, if not add it
            found = await storage.find_all_records(
                type_filter=SCHEMA_SENT_RECORD_TYPE,
                tag_query={
                    "schema_id": schema_id,
                },
            )
            if 0 == len(found):
                await add_schema_non_secrets_record(profile, schema_id)
        except LedgerError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

    if ledger_id:
        return web.json_response({"ledger_id": ledger_id, "schema": schema})
    else:
        return web.json_response({"schema": schema})


def register_events(event_bus: EventBus):
    """Subscribe to any events we need to support."""
    event_bus.subscribe(EVENT_LISTENER_PATTERN, on_schema_event)


async def on_schema_event(profile: Profile, event: Event):
    """Handle any events we need to support."""
    schema_id = event.payload["context"]["schema_id"]

    # after the ledger record is written, write the wallet non-secrets record
    await add_schema_non_secrets_record(profile, schema_id)


async def add_schema_non_secrets_record(profile: Profile, schema_id: str):
    """Write the wallet non-secrets record for a schema (already written to the ledger).

    Args:
        profile: the current profile (used to determine storage)
        schema_id: The schema id (or stringified sequence number)

    """
    schema_id_parts = schema_id.split(":")
    schema_tags = {
        "schema_id": schema_id,
        "schema_issuer_did": schema_id_parts[0],
        "schema_name": schema_id_parts[-2],
        "schema_version": schema_id_parts[-1],
        "epoch": str(int(time())),
    }
    record = StorageRecord(SCHEMA_SENT_RECORD_TYPE, schema_id, schema_tags)
    async with profile.session() as session:
        storage = session.inject(BaseStorage)
        await storage.add_record(record)


async def register(app: web.Application):
    """Register routes."""
    app.add_routes(
        [
            web.post("/schemas", schemas_send_schema),
            web.get("/schemas/created", schemas_created, allow_head=False),
            web.get("/schemas/{schema_id}", schemas_get_schema, allow_head=False),
            web.post(
                "/schemas/{schema_id}/write_record", schemas_fix_schema_wallet_record
            ),
        ]
    )


def post_process_routes(app: web.Application):
    """Amend swagger API."""

    # Add top-level tags description
    if "tags" not in app._state["swagger_dict"]:
        app._state["swagger_dict"]["tags"] = []
    app._state["swagger_dict"]["tags"].append(
        {
            "name": "schema",
            "description": "Schema operations",
            "externalDocs": {
                "description": "Specification",
                "url": (
                    "https://github.com/hyperledger/indy-node/blob/main/"
                    "design/anoncreds.md#schema"
                ),
            },
        }
    )



"""Key DID Resolver.

Resolution is performed using the IndyLedger class.
"""

from typing import Optional, Pattern, Sequence, Text

from ...config.injection_context import InjectionContext
from ...core.profile import Profile
from ...did.did_key import DIDKey
from ...messaging.valid import DIDKey as DIDKeyType
from ..base import BaseDIDResolver, DIDNotFound, ResolverType


class KeyDIDResolver(BaseDIDResolver):
    """Key DID Resolver."""

    def __init__(self):
        """Initialize Key Resolver."""
        super().__init__(ResolverType.NATIVE)

    async def setup(self, context: InjectionContext):
        """Perform required setup for Key DID resolution."""

    @property
    def supported_did_regex(self) -> Pattern:
        """Return supported_did_regex of Key DID Resolver."""
        return DIDKeyType.PATTERN

    async def _resolve(
        self,
        profile: Profile,
        did: str,
        service_accept: Optional[Sequence[Text]] = None,
    ) -> dict:
        """Resolve a Key DID."""
        try:
            did_key = DIDKey.from_did(did)

        except Exception as e:
            raise DIDNotFound(f"Unable to resolve did: {did}") from e

        return did_key.did_doc
